from __future__ import annotations

from dataclasses import dataclass
from dataclasses import field
from functools import partial
import os
import time
from typing import TYPE_CHECKING
from uuid import uuid4

import arrow
import btrfsutil
from btrfsutil import SubvolumeInfo

from btrfs2s3._internal.resolver import Flags
from btrfs2s3._internal.resolver import KeepMeta
from btrfs2s3._internal.resolver import resolve
from btrfs2s3._internal.s3 import iter_backups
from btrfs2s3._internal.thunk import Thunk
from btrfs2s3._internal.thunk import ThunkArg
from btrfs2s3._internal.util import mksubvol
from btrfs2s3._internal.util import SubvolumeFlags

if TYPE_CHECKING:
    from pathlib import Path
    from typing import Sequence

    from mypy_boto3_s3.client import S3Client

    from btrfs2s3._internal import resolver
    from btrfs2s3._internal.action import Actions
    from btrfs2s3._internal.backups import BackupInfo
    from btrfs2s3._internal.preservation import Policy


@dataclass
class SnapshotAssessment:
    initial_path: Path
    info: SubvolumeInfo
    target_path: Thunk[Path]
    real_info: Thunk[SubvolumeInfo]
    keep_meta: KeepMeta = field(default_factory=KeepMeta)

    @property
    def new(self) -> bool:
        return bool(self.keep_meta.flags & Flags.New)


@dataclass
class BackupAssessment:
    backup: Thunk[BackupInfo]
    key: Thunk[str]
    keep_meta: KeepMeta = field(default_factory=KeepMeta)

    @property
    def new(self) -> bool:
        return bool(self.keep_meta.flags & Flags.New)


@dataclass
class SourceAssessment:
    path: Path
    info: SubvolumeInfo
    snapshots: dict[bytes, SnapshotAssessment] = field(default_factory=dict)
    backups: dict[bytes, BackupAssessment] = field(default_factory=dict)


@dataclass(frozen=True)
class Assessment:
    sources: dict[bytes, SourceAssessment] = field(default_factory=dict)


def _get_snapshot_path_for_backup_thunk(
    source: SourceAssessment, backup: Thunk[BackupInfo]
) -> Path:
    return source.snapshots[backup().uuid].target_path()


def _get_send_parent_path_for_backup_thunk(
    source: SourceAssessment, backup: Thunk[BackupInfo]
) -> Path | None:
    info = backup()
    if info.send_parent_uuid is None:
        return None
    return source.snapshots[info.send_parent_uuid].target_path()


def _backup_assessment_to_actions(
    source: SourceAssessment, backup: BackupAssessment, actions: Actions
) -> None:
    if not backup.keep_meta.reasons:
        actions.delete_backup(backup.key)
    elif backup.new:
        snapshot = partial(_get_snapshot_path_for_backup_thunk, source, backup.backup)
        send_parent = partial(
            _get_send_parent_path_for_backup_thunk, source, backup.backup
        )
        actions.create_backup(
            source=source.path,
            snapshot=snapshot,
            send_parent=send_parent,
            key=backup.key,
        )


def _snapshot_assessment_to_actions(
    source: SourceAssessment, snapshot: SnapshotAssessment, actions: Actions
) -> None:
    if not snapshot.keep_meta.reasons:
        actions.delete_snapshot(snapshot.initial_path)
    else:
        if snapshot.new:
            actions.create_snapshot(source=source.path, path=snapshot.initial_path)
        if snapshot.initial_path != snapshot.target_path.peek():
            actions.rename_snapshot(
                source=snapshot.initial_path, target=snapshot.target_path
            )


def assessment_to_actions(assessment: Assessment, actions: Actions) -> None:
    for source in assessment.sources.values():
        for snapshot in source.snapshots.values():
            _snapshot_assessment_to_actions(source, snapshot, actions)
        for backup in source.backups.values():
            _backup_assessment_to_actions(source, backup, actions)


@dataclass
class _SourceAssessor:
    assessment: SourceAssessment
    snapshot_dir: Path
    policy: Policy

    def _make_snapshot_path(self, info: SubvolumeInfo) -> Path:
        ctime = arrow.get(info.ctime, tzinfo=self.policy.tzinfo)
        ctime_str = ctime.isoformat(timespec="seconds")
        name = f"{self.assessment.path.name}.{ctime_str}.{info.ctransid}"
        return self.snapshot_dir / name

    def _make_new_snapshot_path(self) -> Path:
        return self.snapshot_dir / f"{self.assessment.path.name}.NEW.{os.getpid()}"

    def _make_backup_key(self, backup: BackupInfo) -> str:
        suffixes = backup.get_path_suffixes(tzinfo=self.policy.tzinfo)
        return f"{self.assessment.path.name}{''.join(suffixes)}"

    def _is_new_snapshot_needed(self) -> bool:
        if not self.assessment.snapshots:
            return True
        return self.assessment.info.ctransid > max(
            s.info.ctransid for s in self.assessment.snapshots.values()
        )

    def _maybe_propose_new_snapshot(self) -> None:
        if not self._is_new_snapshot_needed():
            return

        proposed_uuid = uuid4().bytes
        now = time.time()
        proposed_info = mksubvol(
            parent_id=self.assessment.info.id,
            flags=SubvolumeFlags.ReadOnly | SubvolumeFlags.Proposed,
            uuid=proposed_uuid,
            parent_uuid=self.assessment.info.uuid,
            generation=self.assessment.info.generation,
            ctransid=self.assessment.info.ctransid,
            otransid=self.assessment.info.generation,
            ctime=now,
            otime=now,
        )
        initial_path = self._make_new_snapshot_path()

        def get_real_info() -> SubvolumeInfo:
            return btrfsutil.subvolume_info(initial_path)

        self.assessment.snapshots[proposed_uuid] = SnapshotAssessment(
            initial_path=initial_path,
            info=proposed_info,
            target_path=Thunk(initial_path),
            real_info=Thunk(get_real_info),
            keep_meta=KeepMeta(flags=Flags.New),
        )

    def _get_target_path(self, snapshot: SnapshotAssessment) -> ThunkArg[Path]:
        if snapshot.real_info.is_tbd():
            return lambda: self._make_snapshot_path(snapshot.real_info())
        return self._make_snapshot_path(snapshot.real_info())

    def _do_resolve(self, *, include_proposed: bool = True) -> resolver.Result:
        snapshots = []
        for snapshot in self.assessment.snapshots.values():
            if snapshot.info.flags & SubvolumeFlags.Proposed and not include_proposed:
                continue
            snapshots.append(snapshot.info)
        backups = [
            b.backup()
            for b in self.assessment.backups.values()
            if not b.backup.is_tbd()
        ]

        return resolve(snapshots=snapshots, backups=backups, policy=self.policy)

    def _resolve(self) -> None:
        # Run resolve including proposed snapshots.
        result = self._do_resolve(include_proposed=True)

        # Mark snapshots as kept, and rename these if necessary
        for uuid, keep_snapshot in result.keep_snapshots.items():
            snapshot = self.assessment.snapshots[uuid]
            snapshot.keep_meta |= keep_snapshot.meta
            # Update the target name, only for snapshots we'll keep
            snapshot.target_path = Thunk(self._get_target_path(snapshot))
        # Mark backups as kept, possibly creating new entries
        for uuid, keep_backup in result.keep_backups.items():
            # NB: if resolve() would ever change BackupInfo in-place, do that
            # here
            if uuid not in self.assessment.backups:
                snapshot = self.assessment.snapshots[uuid]
                if snapshot.real_info.is_tbd():
                    backup = Thunk(
                        partial(
                            lambda p, i: self._get_real_backup(p(), i()),
                            snapshot.target_path,
                            snapshot.real_info,
                        )
                    )
                    key = Thunk(partial(lambda b: self._make_backup_key(b()), backup))
                else:
                    backup = Thunk(keep_backup.item)
                    key = Thunk(self._make_backup_key(keep_backup.item))
                self.assessment.backups[uuid] = BackupAssessment(backup=backup, key=key)
            self.assessment.backups[uuid].keep_meta |= keep_backup.meta

    def _get_real_backup(
        self, target_path: Path, real_info: SubvolumeInfo
    ) -> BackupInfo:
        self.assessment.snapshots[real_info.uuid] = SnapshotAssessment(
            initial_path=target_path,
            info=real_info,
            real_info=Thunk(real_info),
            target_path=Thunk(target_path),
        )
        result = self._do_resolve(include_proposed=False)
        return result.keep_backups[real_info.uuid].item

    def assess(self) -> None:
        self._maybe_propose_new_snapshot()
        self._resolve()


@dataclass
class _Assessor:
    snapshot_dir: Path
    sources: Sequence[Path]
    policy: Policy

    _assessment: Assessment = field(init=False, default_factory=Assessment)

    def _collect_sources(self) -> None:
        for source in self.sources:
            info = btrfsutil.subvolume_info(source)
            self._assessment.sources[info.uuid] = SourceAssessment(
                path=source, info=info
            )

    def _collect_snapshots(self) -> None:
        # SubvolumeIterator only works at a subvolume boundary, so search for
        # one
        search_base = self.snapshot_dir
        while (
            not btrfsutil.is_subvolume(search_base)
            and search_base != search_base.parent
        ):
            search_base = search_base.parent
        if search_base == search_base.parent:
            msg = f"no subvolume found. is {self.snapshot_dir} on a btrfs filesystem?"
            raise RuntimeError(msg)
        for name, info in btrfsutil.SubvolumeIterator(search_base, info=True):
            path = search_base / name
            if self.snapshot_dir not in path.parents:
                continue
            if info.parent_uuid not in self._assessment.sources:
                continue
            if not info.flags & SubvolumeFlags.ReadOnly:
                continue
            self._assessment.sources[info.parent_uuid].snapshots[info.uuid] = (
                SnapshotAssessment(
                    initial_path=path,
                    info=info,
                    real_info=Thunk(info),
                    target_path=Thunk(path),
                )
            )

    def _collect_backups(self, s3: S3Client, bucket: str) -> None:
        for obj, backup in iter_backups(s3, bucket):
            if backup.parent_uuid not in self._assessment.sources:
                continue
            self._assessment.sources[backup.parent_uuid].backups[backup.uuid] = (
                BackupAssessment(backup=Thunk(backup), key=Thunk(obj["Key"]))
            )

    def _assess_for_all_sources(self) -> None:
        for source in self._assessment.sources.values():
            assessor = _SourceAssessor(
                assessment=source, snapshot_dir=self.snapshot_dir, policy=self.policy
            )
            assessor.assess()

    def assess(self, s3: S3Client, bucket: str) -> None:
        self._collect_sources()
        self._collect_snapshots()
        self._collect_backups(s3, bucket)
        self._assess_for_all_sources()

    def get_assessment(self) -> Assessment:
        return self._assessment


def assess(
    *,
    snapshot_dir: Path,
    sources: Sequence[Path],
    s3: S3Client,
    bucket: str,
    policy: Policy,
) -> Assessment:
    assessor = _Assessor(snapshot_dir=snapshot_dir, sources=sources, policy=policy)
    assessor.assess(s3, bucket)
    return assessor.get_assessment()



"""Utilites for manipulating Arrow datetime objects.

The third-party Arrow project implements a feature-rich replacement for the
builtin date and time types. The btrfs2s3 CLI uses this for creating
human-friendly backup schedules.
"""

from __future__ import annotations

from typing import Literal
from typing import TYPE_CHECKING

from typing_extensions import TypeAlias

if TYPE_CHECKING:
    from typing import Iterable

    from arrow import Arrow

    _Bounds: TypeAlias = Literal["[)", "()", "(]", "[]"]


def iter_intersecting_time_spans(
    a: Arrow, *, bounds: _Bounds = "[)"
) -> Iterable[tuple[Arrow, Arrow]]:
    """Convenience function to generate consistent intersecting time spans.

    This does the same thing as iter_time_spans, but will generate all
    human-friendly time spans of all supported timeframes which contain the
    given point in time.

    Args:
        a: A given point in time.
        bounds: Whether the generated time spans should contain their
            endpoints. See the Arrow documentation for more.

    Yields:
        2-tuples of Arrow objects.
    """
    return iter_time_spans(
        a,
        years=(0,),
        quarters=(0,),
        months=(0,),
        weeks=(0,),
        days=(0,),
        hours=(0,),
        minutes=(0,),
        seconds=(0,),
        bounds=bounds,
    )


def iter_time_spans(
    a: Arrow,
    *,
    bounds: _Bounds = "[)",
    years: Iterable[int] = (),
    quarters: Iterable[int] = (),
    months: Iterable[int] = (),
    weeks: Iterable[int] = (),
    days: Iterable[int] = (),
    hours: Iterable[int] = (),
    minutes: Iterable[int] = (),
    seconds: Iterable[int] = (),
) -> Iterable[tuple[Arrow, Arrow]]:
    """Convenience function to generate consistent time spans of multiple timeframes.

    In the Arrow parlance, a time span is a human-friendly interval (e.g. "the
    year 2006") and a timeframe is the type of such an interval (e.g. a year
    versus a week).

    This function generates time spans starting at a given origin time, offset
    by given amounts. For instance,
    iter_time_spans(arrow.get("2006"), years=[-1, 0, 1]) will
    generate three 1-year time spans: the year 2005, the year 2006, and the
    year 2007. The "bounds" specification will be applied consistently to all
    returned time spans.

    Args:
        a: The origin time.
        bounds: Whether the generated time spans should contain their
            endpoints. See the Arrow documentation for more.
        years: An Iterable of offsets of year time spans.
        quarters: An Iterable of offsets of quarter time spans.
        months: An Iterable of offsets of month time spans.
        weeks: An Iterable of offsets of week time spans.
        days: An Iterable of offsets of day time spans.
        hours: An Iterable of offsets of hour time spans.
        minutes: An Iterable of offsets of minute time spans.
        seconds: An Iterable of offsets of second time spans.

    Yields:
        2-tuples of Arrow objects.
    """
    for y in years:
        yield a.shift(years=y).span("year", bounds=bounds)
    for q in quarters:
        yield a.shift(quarters=q).span("quarter", bounds=bounds)
    for m in months:
        yield a.shift(months=m).span("month", bounds=bounds)
    for w in weeks:
        yield a.shift(weeks=w).span("week", bounds=bounds)
    for d in days:
        yield a.shift(days=d).span("day", bounds=bounds)
    for h in hours:
        yield a.shift(hours=h).span("hour", bounds=bounds)
    for m in minutes:
        yield a.shift(minutes=m).span("minute", bounds=bounds)
    for s in seconds:
        yield a.shift(seconds=s).span("second", bounds=bounds)


def convert_span(time_span: tuple[Arrow, Arrow]) -> tuple[float, float]:
    start, end = time_span
    return start.timestamp(), end.timestamp()


def humanize_span(time_span: tuple[Arrow, Arrow], bounds: _Bounds = "[)") -> str:
    a, b = time_span
    if (a, b) == a.span("year", bounds=bounds):
        return a.format("YYYY") + "/P1Y"
    if (a, b) == a.span("quarter", bounds=bounds):
        return f"{a.year:04d}-Q{a.quarter}/P1Q"
    if (a, b) == a.span("month", bounds=bounds):
        return a.format("YYYY-MM") + "/P1M"
    if (a, b) == a.span("week", bounds=bounds):
        return f"{a.year:04d}-W{a.week:02d}/P1W"
    if (a, b) == a.span("day", bounds=bounds):
        return a.format("YYYY-MM-DD") + "/P1D"
    if (a, b) == a.span("hour", bounds=bounds):
        return a.format("YYYY-MM-DDTHH") + "/PT1H"
    if (a, b) == a.span("minute", bounds=bounds):
        return a.format("YYYY-MM-DDTHH:mm") + "/PT1M"
    if (a, b) == a.span("second", bounds=bounds):
        return a.format("YYYY-MM-DDTHH:mm:ss") + "/PT1S"
    return a.format("YYYY-MM-DDTHH:mm:ss") + "/" + b.format("YYYY-MM-DDTHH:mm:ss")



"""Functions for interacting with S3."""

from __future__ import annotations

from typing import TYPE_CHECKING

from btrfs2s3._internal.backups import BackupInfo

if TYPE_CHECKING:
    from typing import Iterator

    from mypy_boto3_s3.client import S3Client
    from mypy_boto3_s3.type_defs import ListObjectsV2RequestRequestTypeDef
    from mypy_boto3_s3.type_defs import ObjectTypeDef


def iter_backups(
    client: S3Client, bucket: str
) -> Iterator[tuple[ObjectTypeDef, BackupInfo]]:
    """Find backups in an S3 bucket that were created by btrfs2s3.

    Args:
        client: An S3 client (e.g. created with boto3.client("s3")).
        bucket: The bucket to enumerate.

    Yields:
        Pairs of object info (as returned by ListObjectsV2) and BackupInfo.
    """
    done = False
    continuation_token: str | None = None
    while not done:
        kwargs: ListObjectsV2RequestRequestTypeDef = {"Bucket": bucket}
        if continuation_token is not None:
            kwargs["ContinuationToken"] = continuation_token
        response = client.list_objects_v2(**kwargs)
        for obj in response.get("Contents", []):
            try:
                info = BackupInfo.from_path(obj["Key"])
            except ValueError:
                continue
            yield obj, info
        done = not response["IsTruncated"]
        if not done:
            continuation_token = response["NextContinuationToken"]



from __future__ import annotations

from typing import TYPE_CHECKING
from unittest.mock import Mock

from btrfs2s3._internal.stream_uploader import upload_non_seekable_stream_via_tempfile
import pytest

if TYPE_CHECKING:
    from typing import IO

    from mypy_boto3_s3.client import S3Client


def test_file_is_uploaded(
    s3: S3Client, stream: IO[bytes], bucket: str, part_size: int, stream_data: bytes
) -> None:
    key = "test-key"

    upload_non_seekable_stream_via_tempfile(
        client=s3, stream=stream, bucket=bucket, key=key, part_size=part_size
    )

    assert s3.get_object(Bucket=bucket, Key=key)["Body"].read() == stream_data


class FakeError(Exception):
    pass


def test_multipart_upload_gets_cancelled_on_error(
    s3: S3Client, stream: IO[bytes], bucket: str, part_size: int
) -> None:
    key = "test-key"

    mock_client = Mock(wraps=s3)
    mock_client.put_object.side_effect = FakeError()
    mock_client.upload_part.side_effect = FakeError()

    with pytest.raises(FakeError):
        upload_non_seekable_stream_via_tempfile(
            client=mock_client,
            stream=stream,
            bucket=bucket,
            key=key,
            part_size=part_size,
        )

    assert s3.list_multipart_uploads(Bucket=bucket).get("Uploads", []) == []



from __future__ import annotations

from typing import Protocol
from typing import TYPE_CHECKING
from uuid import uuid4

import arrow
from btrfs2s3._internal.preservation import Policy
from btrfs2s3._internal.resolver import _Resolver
from btrfs2s3._internal.resolver import Flags
from btrfs2s3._internal.resolver import KeepBackup
from btrfs2s3._internal.resolver import KeepMeta
from btrfs2s3._internal.resolver import KeepSnapshot
from btrfs2s3._internal.resolver import Reasons
from btrfs2s3._internal.resolver import Result
from btrfs2s3._internal.util import backup_of_snapshot
from btrfs2s3._internal.util import mksubvol
import pytest

if TYPE_CHECKING:
    from btrfsutil import SubvolumeInfo


@pytest.fixture()
def parent_uuid() -> bytes:
    return uuid4().bytes


class MkSnap(Protocol):
    def __call__(self, *, t: str | None = None, i: int = 0) -> SubvolumeInfo: ...


@pytest.fixture()
def mksnap(parent_uuid: bytes) -> MkSnap:
    def inner(*, t: str | None = None, i: int = 0) -> SubvolumeInfo:
        a = arrow.get() if t is None else arrow.get(t)
        return mksubvol(
            uuid=uuid4().bytes, parent_uuid=parent_uuid, ctime=a.timestamp(), ctransid=i
        )

    return inner


def test_noop() -> None:
    resolver = _Resolver(snapshots=(), backups=(), policy=Policy())

    resolver.keep_most_recent_snapshot()

    assert resolver.get_result() == Result(keep_snapshots={}, keep_backups={})


def test_one_snapshot(mksnap: MkSnap) -> None:
    snapshot = mksnap()
    resolver = _Resolver(snapshots=(snapshot,), backups=(), policy=Policy())

    resolver.keep_most_recent_snapshot()

    expected_backup = backup_of_snapshot(snapshot, send_parent=None)
    assert resolver.get_result() == Result(
        keep_snapshots={
            snapshot.uuid: KeepSnapshot(snapshot, KeepMeta(reasons=Reasons.MostRecent))
        },
        keep_backups={
            expected_backup.uuid: KeepBackup(
                expected_backup, KeepMeta(reasons=Reasons.MostRecent, flags=Flags.New)
            )
        },
    )


def test_multiple_snapshots_keep_most_recent(mksnap: MkSnap) -> None:
    snapshot1 = mksnap(i=1)
    snapshot2 = mksnap(i=2)
    backup1 = backup_of_snapshot(snapshot1, send_parent=None)
    resolver = _Resolver(
        snapshots=(snapshot1, snapshot2), backups=(backup1,), policy=Policy.all()
    )

    resolver.keep_most_recent_snapshot()

    expected_backup = backup_of_snapshot(snapshot2, send_parent=snapshot1)
    assert resolver.get_result() == Result(
        keep_snapshots={
            snapshot2.uuid: KeepSnapshot(
                snapshot2, KeepMeta(reasons=Reasons.MostRecent)
            )
        },
        keep_backups={
            expected_backup.uuid: KeepBackup(
                expected_backup, KeepMeta(reasons=Reasons.MostRecent, flags=Flags.New)
            )
        },
    )



"""A Thunk wraps a value that might be lazily evaluated."""

from __future__ import annotations

import enum
from functools import total_ordering
from typing import Callable
from typing import Generic
from typing import Literal
from typing import overload
from typing import Protocol
from typing import runtime_checkable
from typing import Union

from typing_extensions import TypeAlias
from typing_extensions import TypeVar


class Error(Exception):
    """The top-level class for errors produced by this module."""


class StillTBDError(Error):
    """Thunk.check() was called, but the thunk hasn't been evaluated yet."""


_T = TypeVar("_T")
_T_contra = TypeVar("_T_contra", contravariant=True)


@runtime_checkable
class _SupportsLT(Protocol[_T_contra]):
    def __lt__(self, other: _T_contra) -> bool: ...


class _TbdType(enum.Enum):
    TBD = enum.auto()

    def __str__(self) -> str:
        return "TBD"


TBD = _TbdType.TBD
"""A sentinel value, meaning the real value hasn't been determined yet."""


@total_ordering
class Thunk(Generic[_T]):
    """A Thunk is a wrapper for a value which may or may not be lazily computed.

    A pre-computed Thunk will return its pre-computed value from peek(),
    check() and __call__().

    A lazily-evaluated Thunk will return the TBD sentinel value from peek(),
    and will raise StillTBDError from check(). __call__() will run the lazy
    evaluation function, and save the value. After this the Thunk will behave
    exactly like a pre-computed Thunk. The evaluation function will not be run
    again.

    This is a very rudimentary lazy evaluation wrapper class. Its only goal is
    to facilitate creating a "dry-run plan" for code that does destructive
    changes, and to present that to the user.

    The intended use of Thunk is as a building block to create complex "intent"
    objects, e.g. an intent to create a backup of some data. The various
    parameters (data source, storage location, etc) of this intent can be
    stored as Thunks which may be precomputed
    or lazily evaluated. This intent can be previewed to a user, and optionally
    executed. Using Thunks allows the preview and execution code to be agnostic
    of data dependencies and edge cases. The preview code can just display all
    the precomputed values in the intent object, and the execution code can
    just evaluate the thunks as needed. Note that this means all the complexity
    is pushed to the code that creates the intent objects.
    """

    _value: _T | Literal[_TbdType.TBD]
    _get_value: Callable[[], _T]

    # Without these overloads, mypy can't infer the type of Thunk(other_thunk),
    # presumably it can't figure out if this should match __init__(_T) or
    # __init__(Thunk[_T]).
    @overload
    def __init__(self, other: Thunk[_T]) -> None: ...
    @overload
    def __init__(self, other: Callable[[], _T]) -> None: ...
    @overload
    def __init__(self, other: _T) -> None: ...
    def __init__(self, other: Thunk[_T] | Callable[[], _T] | _T) -> None:
        """Create a thunk.

        Thunk(other_thunk) creates a copy of other_thunk. The result will have
        the same precomputed value and/or evaluation function as the argument
        Thunk.

        Thunk(callable) creates a lazily-evaluated Thunk. The argument must be
        a zero-args callable.

        Thunk(any-other-value) creates a pre-computed Thunk.

        Args:
            other: A zero-args callable, or another Thunk, or any other
                precomputed value.
        """
        if isinstance(other, Thunk):
            self._get_value = other._get_value  # noqa: SLF001
            self._value = other._value  # noqa: SLF001
        elif callable(other):
            self._get_value = other
            self._value = TBD
        else:
            self._get_value = lambda: other  # pragma: no cover
            self._value = other

    def peek(self) -> _T | Literal[_TbdType.TBD]:
        """Returns the precomputed value, or the TBD sentinel value."""
        return self._value

    def check(self) -> _T:
        """Returns the precomputed value, or raises StillTBDError.

        This asserts that the Thunk is pre-computed. It is mostly useful in
        tests.

        Raises:
            StillTBDError: If the thunk is lazy, and hasn't been evaluated yet.
        """
        value = self._value
        if value is TBD:
            raise StillTBDError
        return value

    def is_tbd(self) -> bool:
        """Returns True if the Thunk is lazy and hasn't been evaluated yet."""
        return self._value is TBD

    def __call__(self) -> _T:
        """Returns the value of the Thunk, possibly running its lazy computation.

        Returns:
            The value of the Thunk.
        """
        if self._value is TBD:
            self._value = self._get_value()
        return self._value

    def __repr__(self) -> str:
        """Return a string representation of the Thunk."""
        value = self._value
        if value is TBD:
            return "Thunk(TBD)"
        return f"Thunk({value!r})"

    def __lt__(self, other: Thunk[_T]) -> bool:
        """Returns whether this Thunk should appear before other Thunks.

        The natural ordering of Thunks preserves the natural ordering of their
        values, except that lazy Thunks appear after precomputed Thunks.
        """
        value = self._value
        othervalue = other._value
        if value is not TBD:
            if othervalue is not TBD:
                if not isinstance(value, _SupportsLT):
                    return NotImplemented
                return value.__lt__(othervalue)
            return True
        if othervalue is not TBD:
            return False
        return id(self) < id(other)


ThunkArg: TypeAlias = Union[Thunk[_T], _T, Callable[[], _T]]
"""An alias to allowed types for the argument to Thunk()."""



import os
from pathlib import Path
import subprocess

from btrfs2s3._internal.action import delete_snapshot
import btrfsutil
import pytest


def test_call(btrfs_mountpoint: Path) -> None:
    source = btrfs_mountpoint / "source"
    btrfsutil.create_subvolume(source)
    snapshot = btrfs_mountpoint / "snapshot"
    btrfsutil.create_snapshot(source, snapshot, read_only=True)

    delete_snapshot(snapshot)

    assert not snapshot.exists()


def test_not_a_subvolume(btrfs_mountpoint: Path) -> None:
    source = btrfs_mountpoint / "source"
    source.mkdir()

    with pytest.raises(RuntimeError, match="target isn't a subvolume"):
        delete_snapshot(source)


def test_not_a_snapshot(btrfs_mountpoint: Path) -> None:
    source = btrfs_mountpoint / "source"
    btrfsutil.create_subvolume(source)

    with pytest.raises(RuntimeError, match="target isn't a snapshot"):
        delete_snapshot(source)


def test_not_a_read_only_snapshot(btrfs_mountpoint: Path) -> None:
    source = btrfs_mountpoint / "source"
    btrfsutil.create_subvolume(source)
    snapshot = btrfs_mountpoint / "snapshot"
    btrfsutil.create_snapshot(source, snapshot, read_only=False)

    with pytest.raises(RuntimeError, match="target isn't a read-only snapshot"):
        delete_snapshot(snapshot)


def test_delete_as_normal_user(btrfs_mountpoint: Path) -> None:
    # This test is narrow. We don't have full test coverage for non-root
    # operation currently. See https://github.com/sbrudenell/btrfs2s3/issues/49

    subprocess.check_call(
        ["mount", "-o", "remount,user_subvol_rm_allowed", btrfs_mountpoint]
    )

    uid = 1000
    os.chown(btrfs_mountpoint, uid, 0)
    os.seteuid(1000)
    try:
        source = btrfs_mountpoint / "source"
        btrfsutil.create_subvolume(source)
        snapshot = btrfs_mountpoint / "snapshot"
        btrfsutil.create_snapshot(source, snapshot, read_only=True)
        delete_snapshot(snapshot)
    finally:
        os.seteuid(0)



from __future__ import annotations

from typing import TYPE_CHECKING

from btrfs2s3._internal.stream_uploader import _iter_parts_via_tempfile

if TYPE_CHECKING:
    from typing import IO


def test_iter_parts_via_tempfile(
    stream: IO[bytes], stream_data: bytes, part_size: int
) -> None:
    expected_data = [
        stream_data[i : i + part_size] for i in range(0, len(stream_data), part_size)
    ]

    got_data = []
    for part_file in _iter_parts_via_tempfile(stream, part_size):
        assert part_file.seekable()
        assert part_file.tell() == 0
        got_data.append(part_file.read())

    assert got_data == expected_data



"""Main code for the btrfs2s3 cli."""

from __future__ import annotations

import argparse
import logging
import sys
from typing import TYPE_CHECKING

from rich.logging import RichHandler

from btrfs2s3._internal.commands import update
from btrfs2s3._internal.console import CONSOLE

if TYPE_CHECKING:
    from typing import Sequence

    from rich.console import Console


_DESCRIPTION = """
btrfs2s3 maintains a tree of incremental backups in cloud storage.
"""

_EPILOG = """
For detailed docs and usage, see https://github.com/sbrudenell/btrfs2s3
"""


def main(*, console: Console | None = None, argv: Sequence[str] | None = None) -> int:
    """Main function for btrfs2s3."""
    console = console if console else CONSOLE
    argv = argv if argv is not None else sys.argv[1:]
    parser = argparse.ArgumentParser(description=_DESCRIPTION, epilog=_EPILOG)

    parser.add_argument(
        "-v", "--verbose", action="store_true", help="enable debug logs"
    )

    subparsers = parser.add_subparsers(
        dest="command", required=True, help="subcommand (required)"
    )

    update.add_args(subparsers.add_parser(update.NAME, **update.ARGS))

    args = parser.parse_args(argv)

    logging.basicConfig(
        level="NOTSET" if args.verbose else logging.INFO,
        format="%(message)s",
        datefmt="[%X]",
        handlers=[RichHandler(console=console)],
    )

    if args.command == update.NAME:
        return update.command(console=console, args=args)
    raise NotImplementedError


if __name__ == "__main__":
    raise SystemExit(main())



from __future__ import annotations

from io import StringIO
import os
from pathlib import Path
import subprocess
from subprocess import DEVNULL
from subprocess import PIPE
from subprocess import Popen
import tempfile
from typing import Protocol
from typing import TYPE_CHECKING
from warnings import warn

import boto3
from btrfs2s3._internal.console import THEME
from moto import mock_aws
import pytest
from rich.console import Console

if TYPE_CHECKING:
    from typing import IO
    from typing import Iterator
    from typing import Sequence

    from mypy_boto3_s3.client import S3Client


@pytest.fixture(autouse=True, scope="session")
def _aws_credentials() -> None:
    # Always stub these out for testing
    os.environ["AWS_ACCESS_KEY_ID"] = "testing"
    os.environ["AWS_SECRET_ACCESS_KEY"] = "testing"
    os.environ["AWS_SECURITY_TOKEN"] = "testing"
    os.environ["AWS_SESSION_TOKEN"] = "testing"
    os.environ["AWS_DEFAULT_REGION"] = "us-east-1"


@pytest.fixture()
def _aws(_aws_credentials: None) -> Iterator[None]:
    with mock_aws():
        yield


@pytest.fixture()
def btrfs_mountpoint() -> Iterator[Path]:
    with tempfile.NamedTemporaryFile() as loop_file:
        loop_file.truncate(2**30)
        subprocess.check_call(["mkfs.btrfs", "-q", loop_file.name])
        with tempfile.TemporaryDirectory() as mount_temp_dir:
            subprocess.check_call(["mount", loop_file.name, mount_temp_dir])
            try:
                yield Path(mount_temp_dir)
            finally:
                subprocess.check_call(["umount", mount_temp_dir])


@pytest.fixture()
def ext4_mountpoint() -> Iterator[Path]:
    with tempfile.NamedTemporaryFile() as loop_file:
        loop_file.truncate(2**30)
        subprocess.check_call(["mkfs.ext4", "-q", loop_file.name])
        with tempfile.TemporaryDirectory() as mount_temp_dir:
            subprocess.check_call(["mount", loop_file.name, mount_temp_dir])
            try:
                yield Path(mount_temp_dir)
            finally:
                subprocess.check_call(["umount", mount_temp_dir])


@pytest.fixture()
def s3(_aws: None) -> S3Client:
    return boto3.client("s3")


@pytest.fixture()
def bucket(s3: S3Client) -> str:
    s3.create_bucket(Bucket="test-bucket")
    return "test-bucket"


class DownloadAndPipe(Protocol):
    def __call__(self, key: str, args: Sequence[str | Path]) -> int: ...


@pytest.fixture()
def download_and_pipe(s3: S3Client, bucket: str) -> DownloadAndPipe:
    def inner(key: str, args: Sequence[str | Path]) -> int:
        process = Popen(args, stdin=PIPE, stdout=DEVNULL)
        # https://github.com/python/typeshed/issues/3831
        assert process.stdin is not None
        s3.download_fileobj(bucket, key, process.stdin)
        # download_fileobj doesn't close its target
        process.stdin.close()
        assert process.wait() == 0
        return process.wait()

    return inner


def pytest_addoption(parser: pytest.Parser) -> None:
    parser.addoption("--update-golden", action="store_true")
    parser.addoption("--remove-stale-golden", action="store_true")


@pytest.fixture(scope="session")
def touched_golden(request: pytest.FixtureRequest) -> Iterator[set[Path]]:
    touched: set[Path] = set()
    yield touched
    found = set()
    for root, _, files in os.walk("tests"):
        for file in files:
            path = Path(root) / file
            if path.suffix == ".golden":
                found.add(path)
    stale = found - touched
    if stale:  # pragma: no cover
        if request.config.getoption("--remove-stale-golden"):
            for path in stale:
                path.unlink()
        else:
            warn(
                "possible stale golden files, re-run with "
                f"--remove-stale-golden: {stale}",
                stacklevel=1,
            )


class Goldify(Protocol):
    def __call__(self, value: str) -> None: ...


@pytest.fixture()
def goldify(request: pytest.FixtureRequest, touched_golden: set[Path]) -> Goldify:
    path = Path(request.node.nodeid + ".golden")
    touched_golden.add(path)

    def inner(value: str) -> None:
        if request.config.getoption("--update-golden"):
            path.write_text(value)  # pragma: no cover
        else:
            assert value == path.read_text()

    return inner


class ConsoleFactory(Protocol):
    def __call__(self, file: IO[str] | None = None) -> Console: ...


@pytest.fixture()
def console_factory() -> ConsoleFactory:
    def inner(file: IO[str] | None = None) -> Console:
        return Console(
            file=file, theme=THEME, width=88, height=30, color_system="truecolor"
        )

    return inner


@pytest.fixture()
def goldifyconsole(
    console_factory: ConsoleFactory, goldify: Goldify
) -> Iterator[Console]:
    file = StringIO()
    console = console_factory(file=file)
    yield console
    goldify(file.getvalue())



from btrfs2s3._internal.preservation import Params
import pytest


def test_args() -> None:
    c = Params(
        years=1, quarters=2, months=3, weeks=4, days=5, hours=6, minutes=7, seconds=8
    )

    assert c.years == 1
    assert c.quarters == 2
    assert c.months == 3
    assert c.weeks == 4
    assert c.days == 5
    assert c.hours == 6
    assert c.minutes == 7
    assert c.seconds == 8


def test_args_invalid() -> None:
    with pytest.raises(ValueError, match="can't be negative"):
        Params(years=-1)


def test_parse() -> None:
    c = Params.parse("1y 4q 12m 52w 365d 8760h 525600M 31536000s")

    assert c.years == 1
    assert c.quarters == 4
    assert c.months == 12
    assert c.weeks == 52
    assert c.days == 365
    assert c.hours == 8760
    assert c.minutes == 525600
    assert c.seconds == 31536000


def test_parse_fail() -> None:
    with pytest.raises(ValueError, match="invalid preservation params"):
        Params.parse("invalid")


def test_parse_fail_wrong_format() -> None:
    with pytest.raises(ValueError, match="invalid preservation params"):
        Params.parse("1m 1y")



from __future__ import annotations

import arrow
from btrfs2s3._internal.commands.update import describe_time_span
from btrfs2s3._internal.zoneinfo import get_zoneinfo
import pytest
from rich.text import Span


@pytest.mark.parametrize(
    ("a_str", "b_str", "expected_plain", "expected_spans", "tzname"),
    [
        (a_str, b_str, expected_plain, expected_spans, tzname)
        for tzname in ("UTC", "America/Los_Angeles")
        for a_str, b_str, expected_plain, expected_spans in [
            ("2006", "2007", "2006 yearly", {Span(0, 4, "iso8601.date")}),
            ("2006", "2006-04-01", "2006-Q1 quarterly", {Span(0, 7, "iso8601.date")}),
            ("2006", "2006-02-01", "2006-01 monthly", {Span(0, 7, "iso8601.date")}),
            (
                "2006-01-02",
                "2006-01-09",
                "2006-W01 weekly",
                {Span(0, 8, "iso8601.date")},
            ),
            (
                "2006-01-02",
                "2006-01-03",
                "2006-01-02 daily",
                {Span(0, 10, "iso8601.date")},
            ),
            (
                "2006-01-02 15",
                "2006-01-02 16",
                "2006-01-02T15 hourly",
                {Span(0, 10, "iso8601.date"), Span(11, 13, "iso8601.time")},
            ),
            (
                "2006-01-02 15:04",
                "2006-01-02 15:05",
                "2006-01-02T15:04 minutely",
                {Span(0, 10, "iso8601.date"), Span(11, 16, "iso8601.time")},
            ),
            (
                "2006-01-02 15:04:05",
                "2006-01-02 15:04:06",
                "2006-01-02T15:04:05 secondly",
                {Span(0, 10, "iso8601.date"), Span(11, 19, "iso8601.time")},
            ),
            (
                "2006",
                "2008",
                "2006-01-01T00:00:00/2008-01-01T00:00:00",
                {
                    Span(0, 10, "iso8601.date"),
                    Span(11, 19, "iso8601.time"),
                    Span(20, 30, "iso8601.date"),
                    Span(31, 39, "iso8601.time"),
                },
            ),
        ]
    ],
)
def test_describe_time_span(
    a_str: str, b_str: str, expected_plain: str, expected_spans: set[Span], tzname: str
) -> None:
    tzinfo = get_zoneinfo(tzname)
    a = arrow.get(a_str, tzinfo=tzinfo)
    b = arrow.get(b_str, tzinfo=tzinfo)
    got = describe_time_span((a.timestamp(), b.timestamp()), tzinfo, bounds="[]")
    assert got.plain == expected_plain
    # The rich-provided highlighter returns loads of spans like iso8601.year
    # which aren't used by our theme. Testing all spans is overfitting. Just
    # test the spans that are used by our theme
    assert set(got.spans) >= expected_spans



from __future__ import annotations

from btrfs2s3._internal.resolver import _MarkedItem
from btrfs2s3._internal.resolver import _Marker
from btrfs2s3._internal.resolver import Flags
from btrfs2s3._internal.resolver import KeepMeta
from btrfs2s3._internal.resolver import Reasons
from btrfs2s3._internal.util import mksubvol
from btrfsutil import SubvolumeInfo


def test_empty() -> None:
    marker: _Marker[SubvolumeInfo] = _Marker()
    assert marker.get_result() == {}


def test_keep_reason() -> None:
    marker: _Marker[SubvolumeInfo] = _Marker()
    snapshot = mksubvol()
    with marker.with_reasons(Reasons.Preserved):
        marker.mark(snapshot)
    assert marker.get_result() == {
        snapshot.uuid: _MarkedItem(
            item=snapshot, meta=KeepMeta(reasons=Reasons.Preserved)
        )
    }


def test_keep_reason_and_time_span() -> None:
    marker: _Marker[SubvolumeInfo] = _Marker()
    snapshot = mksubvol()
    time_span = (0.0, 0.0)
    with marker.with_reasons(Reasons.Preserved), marker.with_time_span(time_span):
        marker.mark(snapshot)
    assert marker.get_result() == {
        snapshot.uuid: _MarkedItem(
            item=snapshot,
            meta=KeepMeta(reasons=Reasons.Preserved, time_spans={time_span}),
        )
    }


def test_keep_reason_flag_and_time_span() -> None:
    marker: _Marker[SubvolumeInfo] = _Marker()
    snapshot = mksubvol()
    time_span = (0.0, 0.0)
    with marker.with_reasons(Reasons.Preserved), marker.with_time_span(time_span):
        marker.mark(snapshot, flags=Flags.New)
    assert marker.get_result() == {
        snapshot.uuid: _MarkedItem(
            item=snapshot,
            meta=KeepMeta(
                reasons=Reasons.Preserved, flags=Flags.New, time_spans={time_span}
            ),
        )
    }


def test_keep_for_multiple_reasons() -> None:
    marker: _Marker[SubvolumeInfo] = _Marker()
    snapshot = mksubvol()
    with marker.with_reasons(Reasons.Preserved):
        marker.mark(snapshot)
        with marker.with_reasons(Reasons.MostRecent):
            marker.mark(snapshot)
    assert marker.get_result() == {
        snapshot.uuid: _MarkedItem(
            item=snapshot, meta=KeepMeta(reasons=Reasons.Preserved | Reasons.MostRecent)
        )
    }


def test_keep_with_reason_context() -> None:
    marker: _Marker[SubvolumeInfo] = _Marker()
    snapshot = mksubvol()
    with marker.with_reasons(Reasons.Preserved):
        marker.mark(snapshot)
    with marker.with_reasons(Reasons.MostRecent):
        marker.mark(snapshot)
    assert marker.get_result() == {
        snapshot.uuid: _MarkedItem(
            item=snapshot, meta=KeepMeta(reasons=Reasons.Preserved | Reasons.MostRecent)
        )
    }


def test_keep_with_reason_and_flags() -> None:
    marker: _Marker[SubvolumeInfo] = _Marker()
    snapshot = mksubvol()
    with marker.with_reasons(Reasons.Preserved):
        marker.mark(snapshot)
        marker.mark(snapshot, flags=Flags.ReplacingNewer)
    assert marker.get_result() == {
        snapshot.uuid: _MarkedItem(
            item=snapshot,
            meta=KeepMeta(reasons=Reasons.Preserved, flags=Flags.ReplacingNewer),
        )
    }


def test_keep_with_time_span_context() -> None:
    marker: _Marker[SubvolumeInfo] = _Marker()
    snapshot = mksubvol()
    time_span = (0.0, 0.0)
    with marker.with_reasons(Reasons.Preserved), marker.with_time_span(time_span):
        marker.mark(snapshot)
    assert marker.get_result() == {
        snapshot.uuid: _MarkedItem[SubvolumeInfo](
            item=snapshot,
            meta=KeepMeta(reasons=Reasons.Preserved, time_spans={time_span}),
        )
    }



"""Code related to schedule-based preservation of snapshots of data.

The design of btrfs2s3's preservation schedule is based on the "retention
policies" of
btrbk. See https://digint.ch/btrbk/doc/btrbk.conf.5.html#_retention_policy for
details.

Broadly, we define a set of time spans, and for each one we preserve the first
snapshot that was created in that time span.

The first snapshot of a year is considered the yearly snapshot. The first
snapshot of a month is the monthly snapshot. We say that a
snapshot is "nominal" for the year and/or month, respectively.

btrfs2s3 supports preserving yearly, quarterly, monthly, weekly,
daily, hourly, "minutely" and "secondly" snapshots.

btrfs2s3 only supports time spans (years, months, days, etc) that are
"grid-aligned". That is, days last from midnight to midnight, never from noon
to noon.

In btrfs2s3, the preservation schedule is significant when choosing parents
for incremental backups. If a policy keeps one monthly and one yearly
backup, then we will store the yearly backup as a full backup. We will store the
monthly backup as an incremental
backup, having the yearly full backup as the parent.

The time zone is significant when defining a preservation schedule (for example,
"midnight" corresponds to a different unix timestamp depending on the time
zone). A change to the time zone of a preservation schedule can completely change
the nominality of snapshots, which could lead to undesired deletion of data,
longer-than-expected incremental backup chains, or heavily-duplicated full
backups.

To avoid this chaos, we never honor the system locales or time zones. We require
time  zones to be explicitly specified, defaulting to UTC.

By convention, time spans are represented as
tuples of left-closed right-open intervals (that is, the same convention used by
python slices). For example, the year 2006 in UTC is represented as (1136073600.0,
1167609600.0), even though it does NOT include the upper boundary timestamp of
1167609600.0.
"""

from __future__ import annotations

import dataclasses
from datetime import timezone
from datetime import tzinfo
import re
from typing import Generic
from typing import Literal
from typing import Tuple
from typing import TYPE_CHECKING
from typing import TypeVar

import arrow
from typing_extensions import Self
from typing_extensions import TypeAlias
from typing_extensions import TypedDict

from btrfs2s3._internal.arrowutil import convert_span
from btrfs2s3._internal.arrowutil import iter_time_spans

if TYPE_CHECKING:
    from typing import Iterator

TS: TypeAlias = Tuple[float, float]
"""An alias for tuple[float, float] which is used as a time span type.

This is just provided because it's frequently used, and shorter than
tuple[float, float].
"""

Timeframe: TypeAlias = Literal[
    "seconds", "minutes", "hours", "days", "weeks", "months", "quarters", "years"
]
"""A type alias for the valid arguments to Params."""
TIMEFRAMES: frozenset[Timeframe] = frozenset(
    ("seconds", "minutes", "hours", "days", "weeks", "months", "quarters", "years")
)
"""A list of valid arguments to Params."""

_T = TypeVar("_T")


class TimeframeArgs(TypedDict, Generic[_T], total=False):
    """A TypedDict for the valid arguments to Params."""

    years: _T
    quarters: _T
    months: _T
    weeks: _T
    days: _T
    hours: _T
    minutes: _T
    seconds: _T


_REGEX = re.compile(
    r"^(\b(?P<years>\d+)y\b)? ??"
    r"(\b(?P<quarters>\d+)q\b)? ??"
    r"(\b(?P<months>\d+)m\b)? ??"
    r"(\b(?P<weeks>\d+)w\b)? ??"
    r"(\b(?P<days>\d+)d\b)? ??"
    r"(\b(?P<hours>\d+)h\b)? ??"
    r"(\b(?P<minutes>\d+)M\b)? ??"
    r"(\b(?P<seconds>\d+)s\b)? ??$"
)


@dataclasses.dataclass
class Params:
    """Parameters which define a schedule of preserving snapshots or backups.

    These parameters have the same significance as in btrbk's retention
    policies. See
    https://digint.ch/btrbk/doc/btrbk.conf.5.html#_retention_policy for more.

    If any of the attributes is zero, then snapshots/backups will not be
    preserved on that interval. None of the attributes may be negative.

    Attributes:
        years: How many years back yearly backups should be preserved. The
            first backup in a year is the yearly backup.
        quarters: How many quarters back quarterly backups should be preserved.
            The first backup in a quarter is the quarterly backup.
        months: How many months back monthly backups should be preserved. The
            first backup in a month is the monthly backup.
        weeks: How many weeks back weekly backups should be preserved. The
            first backup in a week is the weekly backup.
        days: How many days back daily backups should be preserved. The first
            backup in a day is the daily backup.
        hours: How many hours back hourly backups should be preserved. The
            first backup in an hour is the hourly backup.
        minutes: How many minutes back "minutely" backups should be preserved.
            The first backup in a minute is the minutely backup.
        seconds: How many seconds back "secondly" backups should be preserved.
            The first backup in a second is the secondly backup.
    """

    years: int = 0
    quarters: int = 0
    months: int = 0
    weeks: int = 0
    days: int = 0
    hours: int = 0
    minutes: int = 0
    seconds: int = 0

    def __post_init__(self) -> None:
        """Do validation checks."""
        for arg in TIMEFRAMES:
            if getattr(self, arg) < 0:
                msg = f"parameter can't be negative: {arg}={getattr(self, arg)}"
                raise ValueError(msg)

    @classmethod
    def all(cls) -> Self:
        """Returns a Params which preserves ALL snapshots or backups.

        This is mainly useful for testing.

        It's also not currently correct, until we implement infinite
        preservation.
        """
        kwargs = {t: 1 for t in TIMEFRAMES}
        # https://github.com/python/mypy/issues/10023
        return cls(**kwargs)  # type: ignore[misc]

    @classmethod
    def parse(cls, desc: str) -> Self:
        """Parse a Params from a human-friendly description string.

        The format for the description string is based on the one used by
        btrbk. See
        https://digint.ch/btrbk/doc/btrbk.conf.5.html#_retention_policy

        The format is:

            [<yearly>y] [<quarterly>q] [<monthly>m] [<weekly>w] [<daily>d]
            [<hourly>h] [<minutely>M] [<secondly>s]

        Args:
            desc: The description string.
        """
        m = _REGEX.match(desc)
        if not m:
            msg = f"invalid preservation params: {desc}"
            raise ValueError(msg)
        kwargs = TimeframeArgs[int]()
        for timeframe in TIMEFRAMES:
            value = m.group(timeframe)
            if value is not None:
                kwargs[timeframe] = int(value)
        return cls(**kwargs)


class Policy:
    """A schedule-based policy for preserving snapshots.

    A Policy is a helper class which defines a schedule for preserving
    snapshots (or backups) of data.

    The preservation schedule is defined by Params, which are only meaningful
    relative to the "current" time. For consistency, a Policy object fixes the
    current time when it is created (the value can be supplied in the
    constructor). The return value of should_preserve_for_time_span() is considered
    relative to this creation time, not the time the function is called.
    """

    @classmethod
    def all(cls, *, tzinfo: tzinfo | None = None, now: float | None = None) -> Self:
        """Returns a Policy which preserves ALL snapshots or backups.

        This is mainly useful for testing (and not currently correct, until we
        implement infinite preservation).
        """
        return cls(params=Params.all(), tzinfo=tzinfo, now=now)

    def __init__(
        self,
        *,
        params: Params | None = None,
        tzinfo: tzinfo | None = None,
        now: float | None = None,
    ) -> None:
        """Creates a Policy.

        When no args are supplied, the resulting Policy will not
        preserve anything. This can be useful to bypass schedule-based
        preservation logic.

        Args:
            params: The parameters which define the schedule of time spans for
                which backups should be preserved.
            tzinfo: The time zone in which to operate. This will change the
                boundaries of all time spans. If None, defaults to UTC.
            now: A timestamp to use as the current time. If None, defaults to
                the actual current time. This is mainly useful for testing.
        """
        if params is None:
            params = Params()
        if tzinfo is None:
            tzinfo = timezone.utc
        if now is None:
            now = arrow.get(tzinfo=tzinfo).timestamp()

        self._params = params
        self._tzinfo = tzinfo
        self._now = now

        kwargs = {t: range(0, -getattr(params, t), -1) for t in TIMEFRAMES}
        # https://github.com/python/mypy/issues/10023
        self._preserve_for_time_spans = {
            convert_span(span)
            for span in iter_time_spans(
                arrow.get(now, tzinfo=self._tzinfo),
                bounds="[]",
                **kwargs,  # type: ignore[misc]
            )
        }

    @property
    def tzinfo(self) -> tzinfo:
        """Returns the time zone object used by this Policy."""
        return self._tzinfo

    @property
    def now(self) -> float:
        """Returns the current time used by this Policy."""
        return self._now

    def iter_time_spans(self, timestamp: float) -> Iterator[TS]:
        """Yields time spans which overlap with a given timestamp.

        This can be used to determine which time spans a snapshot falls into. A
        caller can then call should_preserve_for_time_span() to figure out whether the
        snapshot should be retained.

        Note that this only yields time spans which overlap the argument
        timestamp AND may EVER be relevant to this policy (otherwise, this could
        just be a top-level function). If this policy only preserves monthly backups,
        this function will only yield monthly time spans, and never yearly
        ones. This is significant for the function of btrfs2s3.

        btrfs2s3 (in resolver.py) will store backups as full or incremental
        based primarily on this function. If a backup is nominal for the FIRST
        time span returned by this function, it will be stored as a full
        backup. Otherwise, it will be stored as an incremental backup, and its
        parent will be the nominal backup of the next-earliest time span
        returned by this function for which the backup in question is NOT
        nominal. For example, if a Policy calls for yearly and monthly
        backups, then the yearly backups will be full backups (because the
        yearly time spans are returned first by iter_time_spans()) and the
        monthly backups will be incremental backups with a yearly backup as
        their parent (the monthly time spans are returned second by this
        function, so btrfs2s3 will look at yearly intervals which are returned
        earlier, and find the nominal backup for that interval to use as the
        parent).

        Args:
            timestamp: A timestamp to query.

        Yields:
            (timestamp, timestamp) time span tuples, in descending order of
                length.
        """
        kwargs = {t: (0,) if getattr(self._params, t) else () for t in TIMEFRAMES}
        # https://github.com/python/mypy/issues/10023
        for span in iter_time_spans(
            arrow.get(timestamp, tzinfo=self._tzinfo),
            bounds="[]",
            **kwargs,  # type: ignore[misc]
        ):
            yield convert_span(span)

    def should_preserve_for_time_span(self, time_span: TS) -> bool:
        """Returns whether we want to retain a snapshot for a time span.

        The return value is meaningful relative to the Policy's creation time
        (policy.now), not the current time when the function is called.

        Args:
            time_span: The (timestamp, timestamp) time span tuple in question.

        Returns:
            Whether this policy calls for a snapshot to be retained for the
                given time span.
        """
        return time_span in self._preserve_for_time_spans



from __future__ import annotations

from enum import IntFlag

from btrfsutil import SubvolumeInfo

from btrfs2s3._internal.backups import BackupInfo

NULL_UUID = b"\0" * 16


def mksubvol(
    *,
    id: int = 0,  # noqa: A002
    parent_id: int = 0,
    dir_id: int = 0,
    flags: int = 0,
    uuid: bytes = NULL_UUID,
    parent_uuid: bytes = NULL_UUID,
    received_uuid: bytes = NULL_UUID,
    generation: int = 0,
    ctransid: int = 0,
    otransid: int = 0,
    stransid: int = 0,
    rtransid: int = 0,
    ctime: float = 0.0,
    otime: float = 0.0,
    stime: float = 0.0,
    rtime: float = 0.0,
) -> SubvolumeInfo:
    return SubvolumeInfo(
        (
            id,
            parent_id,
            dir_id,
            flags,
            uuid,
            parent_uuid,
            received_uuid,
            generation,
            ctransid,
            otransid,
            stransid,
            rtransid,
            ctime,
            otime,
        ),
        {"stime": stime, "rtime": rtime},
    )


def backup_of_snapshot(
    snapshot: SubvolumeInfo, send_parent: SubvolumeInfo | None = None
) -> BackupInfo:
    return BackupInfo(
        uuid=snapshot.uuid,
        parent_uuid=snapshot.parent_uuid,
        send_parent_uuid=None if send_parent is None else send_parent.uuid,
        ctransid=snapshot.ctransid,
        ctime=snapshot.ctime,
    )


class SubvolumeFlags(IntFlag):
    ReadOnly = 1
    Proposed = 2**30



from pathlib import Path

from btrfs2s3._internal.action import rename_snapshot
import btrfsutil


def test_call(btrfs_mountpoint: Path) -> None:
    source = btrfs_mountpoint / "source"
    btrfsutil.create_subvolume(source)
    snapshot = btrfs_mountpoint / "snapshot"
    btrfsutil.create_snapshot(source, snapshot, read_only=True)

    target = btrfs_mountpoint / "new-snapshot"

    rename_snapshot(source=snapshot, target=target)

    assert not snapshot.exists()
    assert target.exists()



from __future__ import annotations

from typing import Protocol
from typing import TYPE_CHECKING

import arrow
from btrfs2s3._internal.preservation import Params
from btrfs2s3._internal.preservation import Policy
from btrfs2s3._internal.resolver import Flags
from btrfs2s3._internal.resolver import KeepBackup
from btrfs2s3._internal.resolver import KeepMeta
from btrfs2s3._internal.resolver import KeepSnapshot
from btrfs2s3._internal.resolver import Reasons
from btrfs2s3._internal.resolver import resolve
from btrfs2s3._internal.resolver import Result
from btrfs2s3._internal.util import backup_of_snapshot
from btrfs2s3._internal.util import mksubvol
import pytest

if TYPE_CHECKING:
    from btrfsutil import SubvolumeInfo

from uuid import uuid4


@pytest.fixture()
def parent_uuid() -> bytes:
    return uuid4().bytes


class MkSnap(Protocol):
    def __call__(self, *, t: str | None = None, i: int = 0) -> SubvolumeInfo: ...


@pytest.fixture()
def mksnap(parent_uuid: bytes) -> MkSnap:
    def inner(*, t: str | None = None, i: int = 0) -> SubvolumeInfo:
        a = arrow.get() if t is None else arrow.get(t)
        return mksubvol(
            uuid=uuid4().bytes, parent_uuid=parent_uuid, ctime=a.timestamp(), ctransid=i
        )

    return inner


def _t(t: str) -> float:
    return arrow.get(t).timestamp()


def test_noop() -> None:
    result = resolve(snapshots=(), backups=(), policy=Policy())

    assert result == Result(keep_snapshots={}, keep_backups={})


def test_one_snapshot_preserved(mksnap: MkSnap) -> None:
    snapshot = mksnap(t="2006-01-01", i=1)
    result = resolve(
        snapshots=(snapshot,),
        backups=(),
        policy=Policy(now=arrow.get("2006-01-01").timestamp(), params=Params(years=1)),
    )

    expected_backup = backup_of_snapshot(snapshot, send_parent=None)
    assert result == Result(
        keep_snapshots={
            snapshot.uuid: KeepSnapshot(
                snapshot,
                KeepMeta(
                    reasons=Reasons.Preserved | Reasons.MostRecent,
                    time_spans={(_t("2006-01-01"), _t("2007-01-01"))},
                ),
            )
        },
        keep_backups={
            expected_backup.uuid: KeepBackup(
                expected_backup,
                KeepMeta(
                    reasons=Reasons.Preserved | Reasons.MostRecent,
                    flags=Flags.New,
                    time_spans={(_t("2006-01-01"), _t("2007-01-01"))},
                ),
            )
        },
    )


def test_multiple_snapshots_and_time_spans(mksnap: MkSnap) -> None:
    snapshot1 = mksnap(t="2006-01-01", i=1)
    snapshot2 = mksnap(t="2006-01-02", i=2)
    snapshot3 = mksnap(t="2006-01-02", i=3)
    result = resolve(
        snapshots=(snapshot1, snapshot2, snapshot3),
        backups=(),
        policy=Policy(
            now=arrow.get("2006-01-02").timestamp(), params=Params(years=1, months=1)
        ),
    )

    expected_backup1 = backup_of_snapshot(snapshot1, send_parent=None)
    expected_backup3 = backup_of_snapshot(snapshot3, send_parent=snapshot1)
    assert result == Result(
        keep_snapshots={
            snapshot1.uuid: KeepSnapshot(
                snapshot1,
                KeepMeta(
                    reasons=Reasons.Preserved,
                    time_spans={
                        (_t("2006-01-01"), _t("2007-01-01")),
                        (_t("2006-01-01"), _t("2006-02-01")),
                    },
                ),
            ),
            snapshot3.uuid: KeepSnapshot(
                snapshot3, KeepMeta(reasons=Reasons.MostRecent)
            ),
        },
        keep_backups={
            expected_backup1.uuid: KeepBackup(
                expected_backup1,
                KeepMeta(
                    reasons=Reasons.Preserved,
                    flags=Flags.New,
                    time_spans={
                        (_t("2006-01-01"), _t("2007-01-01")),
                        (_t("2006-01-01"), _t("2006-02-01")),
                    },
                ),
            ),
            expected_backup3.uuid: KeepBackup(
                expected_backup3, KeepMeta(reasons=Reasons.MostRecent, flags=Flags.New)
            ),
        },
    )


def test_keep_send_ancestor_on_year_change(mksnap: MkSnap) -> None:
    snapshot1 = mksnap(t="2006-01-01", i=1)
    snapshot2 = mksnap(t="2006-12-01", i=2)
    snapshot3 = mksnap(t="2007-01-01", i=3)
    backup1 = backup_of_snapshot(snapshot1, send_parent=None)
    backup2 = backup_of_snapshot(snapshot2, send_parent=snapshot1)
    result = resolve(
        snapshots=(snapshot1, snapshot2, snapshot3),
        backups=(backup1, backup2),
        policy=Policy(
            now=arrow.get("2007-01-01").timestamp(), params=Params(years=1, months=2)
        ),
    )

    expected_backup3 = backup_of_snapshot(snapshot3, send_parent=None)
    assert result == Result(
        keep_snapshots={
            snapshot2.uuid: KeepSnapshot(
                snapshot2,
                KeepMeta(
                    reasons=Reasons.Preserved,
                    time_spans={(_t("2006-12-01"), _t("2007-01-01"))},
                ),
            ),
            snapshot3.uuid: KeepSnapshot(
                snapshot3,
                KeepMeta(
                    reasons=Reasons.Preserved | Reasons.MostRecent,
                    time_spans={
                        (_t("2007-01-01"), _t("2008-01-01")),
                        (_t("2007-01-01"), _t("2007-02-01")),
                    },
                ),
            ),
        },
        keep_backups={
            backup1.uuid: KeepBackup(
                backup1,
                KeepMeta(reasons=Reasons.SendAncestor, other_uuids={backup2.uuid}),
            ),
            backup2.uuid: KeepBackup(
                backup2,
                KeepMeta(
                    reasons=Reasons.Preserved,
                    time_spans={(_t("2006-12-01"), _t("2007-01-01"))},
                ),
            ),
            expected_backup3.uuid: KeepBackup(
                expected_backup3,
                KeepMeta(
                    reasons=Reasons.Preserved | Reasons.MostRecent,
                    flags=Flags.New,
                    time_spans={
                        (_t("2007-01-01"), _t("2008-01-01")),
                        (_t("2007-01-01"), _t("2007-02-01")),
                    },
                ),
            ),
        },
    )


def test_backup_chain_broken(mksnap: MkSnap) -> None:
    snapshot1 = mksnap(t="2005-01-01", i=1)
    snapshot2 = mksnap(t="2006-01-01", i=2)
    backup2 = backup_of_snapshot(snapshot2, send_parent=snapshot1)

    with pytest.warns(UserWarning, match="Backup chain is broken"):
        result = resolve(
            snapshots=(),
            backups=(backup2,),
            policy=Policy(
                now=arrow.get("2006-01-01").timestamp(), params=Params(years=1)
            ),
        )

    assert result == Result(
        keep_snapshots={},
        keep_backups={
            backup2.uuid: KeepBackup(
                backup2,
                KeepMeta(
                    reasons=Reasons.Preserved,
                    flags=Flags.NoSnapshot,
                    time_spans={(_t("2006-01-01"), _t("2007-01-01"))},
                ),
            )
        },
    )



from __future__ import annotations

import itertools
import random
import re
from typing import TYPE_CHECKING
from uuid import UUID

import arrow
from btrfs2s3._internal.backups import BackupInfo
import pytest

if TYPE_CHECKING:
    from typing import Sequence
    from typing import TypeVar

    _T = TypeVar("_T")


def test_get_path_suffixes_with_real_timezone() -> None:
    info = BackupInfo(
        uuid=UUID("3fd11d8e-8110-4cd0-b85c-bae3dda86a3d").bytes,
        parent_uuid=UUID("9d9d3bcb-4b62-46a3-b6e2-678eeb24f54e").bytes,
        ctransid=12345,
        ctime=arrow.get("2006-01-01", tzinfo="US/Pacific").timestamp(),
        send_parent_uuid=UUID("3ae01eae-d50d-4187-b67f-cef0ef973e1f").bytes,
    )
    got = info.get_path_suffixes(tzinfo="US/Pacific")
    expected = [
        ".ctim2006-01-01T00:00:00-08:00",
        ".ctid12345",
        ".uuid3fd11d8e-8110-4cd0-b85c-bae3dda86a3d",
        ".sndp3ae01eae-d50d-4187-b67f-cef0ef973e1f",
        ".prnt9d9d3bcb-4b62-46a3-b6e2-678eeb24f54e",
        ".mdvn1",
        ".seqn0",
    ]
    assert got == expected

    round_trip = BackupInfo.from_path(f"name{''.join(got)}.gz")
    assert round_trip == info


def test_get_path_suffixes_default_to_utc() -> None:
    info = BackupInfo(
        uuid=UUID("3fd11d8e-8110-4cd0-b85c-bae3dda86a3d").bytes,
        parent_uuid=UUID("9d9d3bcb-4b62-46a3-b6e2-678eeb24f54e").bytes,
        ctransid=12345,
        ctime=arrow.get("2006-01-01").timestamp(),
        send_parent_uuid=UUID("3ae01eae-d50d-4187-b67f-cef0ef973e1f").bytes,
    )
    got = info.get_path_suffixes()
    expected = [
        ".ctim2006-01-01T00:00:00+00:00",
        ".ctid12345",
        ".uuid3fd11d8e-8110-4cd0-b85c-bae3dda86a3d",
        ".sndp3ae01eae-d50d-4187-b67f-cef0ef973e1f",
        ".prnt9d9d3bcb-4b62-46a3-b6e2-678eeb24f54e",
        ".mdvn1",
        ".seqn0",
    ]
    assert got == expected

    round_trip = BackupInfo.from_path(f"name{''.join(got)}.gz")
    assert round_trip == info


def test_get_path_suffixes_with_full_backup() -> None:
    info = BackupInfo(
        uuid=UUID("3fd11d8e-8110-4cd0-b85c-bae3dda86a3d").bytes,
        parent_uuid=UUID("9d9d3bcb-4b62-46a3-b6e2-678eeb24f54e").bytes,
        ctransid=12345,
        ctime=arrow.get("2006-01-01").timestamp(),
        send_parent_uuid=None,
    )
    got = info.get_path_suffixes()
    expected = [
        ".ctim2006-01-01T00:00:00+00:00",
        ".ctid12345",
        ".uuid3fd11d8e-8110-4cd0-b85c-bae3dda86a3d",
        ".sndp00000000-0000-0000-0000-000000000000",
        ".prnt9d9d3bcb-4b62-46a3-b6e2-678eeb24f54e",
        ".mdvn1",
        ".seqn0",
    ]
    assert got == expected

    round_trip = BackupInfo.from_path(f"name{''.join(got)}.gz")
    assert round_trip == info


def _fixed_choices(population: Sequence[_T], seed: int, k: int) -> list[_T]:
    state = random.getstate()
    random.seed(seed)
    result = random.choices(population, k=k)
    random.setstate(state)
    return result


@pytest.mark.parametrize(
    "suffixes",
    _fixed_choices(
        list(
            itertools.permutations(
                [
                    ".ctim2006-01-01T00:00:00-08:00",
                    ".ctid12345",
                    ".uuid3fd11d8e-8110-4cd0-b85c-bae3dda86a3d",
                    ".sndp3ae01eae-d50d-4187-b67f-cef0ef973e1f",
                    ".prnt9d9d3bcb-4b62-46a3-b6e2-678eeb24f54e",
                    ".mdvn1",
                    ".seqn0",
                    ".gz",
                ]
            )
        ),
        0,
        1000,
    ),
)
def test_from_path_with_suffixes_in_any_order(suffixes: Sequence[str]) -> None:
    path = f"my-backup{''.join(suffixes)}"
    got = BackupInfo.from_path(path)
    expected = BackupInfo(
        uuid=UUID("3fd11d8e-8110-4cd0-b85c-bae3dda86a3d").bytes,
        parent_uuid=UUID("9d9d3bcb-4b62-46a3-b6e2-678eeb24f54e").bytes,
        ctransid=12345,
        ctime=arrow.get("2006-01-01", tzinfo="US/Pacific").timestamp(),
        send_parent_uuid=UUID("3ae01eae-d50d-4187-b67f-cef0ef973e1f").bytes,
    )
    assert got == expected


@pytest.mark.parametrize(
    "bad_path",
    [
        ".ctim20O6-01-01T00:00:00-08:00"
        ".ctid12345"
        ".uuid3fd11d8e-8110-4cd0-b85c-bae3dda86a3d"
        ".sndp3ae01eae-d50d-4187-b67f-cef0ef973e1f"
        ".prnt9d9d3bcb-4b62-46a3-b6e2-678eeb24f54e"
        ".mdvn1"
        ".seqn0",
        ".ctim2006-01-01T00:00:00-08:00"
        ".ctid12345"
        ".uuid3fd11d8e-811O-4cd0-b85c-bae3dda86a3d"
        ".sndp3ae01eae-d50d-4187-b67f-cef0ef973e1f"
        ".prnt9d9d3bcb-4b62-46a3-b6e2-678eeb24f54e"
        ".mdvn1"
        ".seqn0",
        ".ctim2006-01-01T00:00:00-08:00"
        ".ctidl2345.u3fd11d8e-8110-4cd0-b85c-bae3dda86a3d"
        ".sndp3ae01eae-d50d-4187-b67f-cef0ef973e1f"
        ".prnt9d9d3bcb-4b62-46a3-b6e2-678eeb24f54e"
        ".mdvn1"
        ".seqn0",
        ".ctim2006-01-01T00:00:00-08:00"
        ".ctid12345"
        ".uuid3fd11d8e-8110-4cd0-b85c-bae3dda86a3d"
        ".sndp3ae01eae-d50d-4187-b67f-cef0ef973e1f"
        ".prnt9d9d3bcb-4b62-46a3-b6e2-678eeb24f54e"
        ".mdvn1"
        ".seqn0",
        ".ctim2006-01-01T00:00:00-08:00"
        ".ctid12345"
        ".uuid3fd11d8e-8110-4cd0-b85c-bae3dda86a3d"
        ".sndp3ae01eae-d50d-4187-b67f-cef0ef973e1f"
        ".prnt9d9d3gcb-4b62-46a3-b6e2-678eeb24f54e"
        ".mdvn1"
        ".seqn0",
    ],
)
def test_bad_paths(bad_path: str) -> None:
    with pytest.raises(
        ValueError, match="missing or incomplete parameters for backup name"
    ):
        BackupInfo.from_path(bad_path)


@pytest.mark.parametrize(
    "bad_path",
    [
        "bad-path-no-suffixes",
        "bad.path.with.suffixes",
        ".ctim2006-01-01T00:00:00-08:00"
        ".ctid12345"
        ".uuid3fd11d8e-8110-4cd0-b85c-bae3dda86a3d"
        ".sndp3ae01eae-d50d-4187-b67f-cef0ef973e1f"
        ".prnt9d9d3bcb-4b62-46a3-b6e2-678eeb24f54e"
        ".seqn0",
        ".ctim2006-01-01T00:00:00-08:00"
        ".ctid12345"
        ".uuid3fd11d8e-8110-4cd0-b85c-bae3dda86a3d"
        ".sndp3ae01eae-d50d-4187-b67f-cef0ef973e1f"
        ".prnt9d9d3bcb-4b62-46a3-b6e2-678eeb24f54e"
        ".mdvnI"
        ".seqn0",
    ],
)
def test_no_version(bad_path: str) -> None:
    with pytest.raises(
        ValueError,
        match=re.escape("backup name metadata version missing (not a backup?)"),
    ):
        BackupInfo.from_path(bad_path)


def test_bad_version() -> None:
    with pytest.raises(ValueError, match="unsupported backup name metadata version"):
        BackupInfo.from_path(
            ".ctim2006-01-01T00:00:00-08:00"
            ".ctid12345"
            ".uuid3fd11d8e-8110-4cd0-b85c-bae3dda86a3d"
            ".sndp3ae01eae-d50d-4187-b67f-cef0ef973e1f"
            ".prnt9d9d3bcb-4b62-46a3-b6e2-678eeb24f54e"
            ".mdvn1000"
            ".seqn0"
        )


def test_bad_sequence_number() -> None:
    with pytest.raises(ValueError, match="unsupported sequence number"):
        BackupInfo.from_path(
            ".ctim2006-01-01T00:00:00-08:00"
            ".ctid12345"
            ".uuid3fd11d8e-8110-4cd0-b85c-bae3dda86a3d"
            ".sndp3ae01eae-d50d-4187-b67f-cef0ef973e1f"
            ".prnt9d9d3bcb-4b62-46a3-b6e2-678eeb24f54e"
            ".mdvn1"
            ".seqn1"
        )



from __future__ import annotations

import os
from tempfile import TemporaryFile
from threading import Thread
from typing import cast
from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from typing import IO
    from typing import Iterator


@pytest.fixture(params=[5 * 2**20, 5 * 2**20 + 512, 10 * 2**20])
def data_size(request: pytest.FixtureRequest) -> int:
    return cast(int, request.param)


@pytest.fixture(params=[5 * 2**20, 5 * 2**20 + 512, 10 * 2**20])
def part_size(request: pytest.FixtureRequest) -> int:
    return cast(int, request.param)


@pytest.fixture()
def stream_data(data_size: int) -> bytes:
    return os.urandom(data_size)


@pytest.fixture(params=[-1, 0], ids=["buffered", "unbuffered"])
def buffering(request: pytest.FixtureRequest) -> int:
    return cast(int, request.param)


@pytest.fixture(params=[False, True], ids=["nonseekable", "seekable"])
def seekable(request: pytest.FixtureRequest) -> bool:
    return cast(bool, request.param)


@pytest.fixture()
def stream(buffering: int, seekable: bool, stream_data: bytes) -> Iterator[IO[bytes]]:  # noqa: FBT001
    if seekable:
        with TemporaryFile(buffering=buffering) as stream:
            stream.write(stream_data)
            stream.seek(0, os.SEEK_SET)
            yield stream
    else:
        read_fd, write_fd = os.pipe()

        def fill(write_fd: int) -> None:
            os.write(write_fd, stream_data)
            os.close(write_fd)

        fill_thread = Thread(target=fill, args=(write_fd,))
        fill_thread.start()
        with open(read_fd, mode="rb", buffering=buffering) as stream:  # noqa: PTH123
            yield stream
        fill_thread.join()



"""Types and functions relating to backups."""

from __future__ import annotations

from contextlib import suppress
import dataclasses
from math import floor
import pathlib
from typing import TYPE_CHECKING
from uuid import UUID

import arrow
from arrow import Arrow
from typing_extensions import Self

if TYPE_CHECKING:
    from datetime import tzinfo
    from typing import Sequence

_CTIM = "ctim"
_CTID = "ctid"
_UUID = "uuid"
_SNDP = "sndp"
_PRNT = "prnt"
_MDVN = "mdvn"
_SEQN = "seqn"

_METADATA_VERSION = 1


@dataclasses.dataclass(frozen=True)
class BackupInfo:
    """Information about a backup."""

    uuid: bytes
    parent_uuid: bytes
    send_parent_uuid: bytes | None
    ctransid: int
    ctime: float

    def __post_init__(self) -> None:
        """Post-initialization fixups."""
        # We need to use object.__setattr__ to fix up an attribute for a frozen
        # instance. Not sure if there's a better way to do this
        object.__setattr__(self, "ctime", floor(self.ctime))

    def get_path_suffixes(self, *, tzinfo: tzinfo | str | None = None) -> Sequence[str]:
        """Get path suffixes suitable for naming a backup as an S3 object.

        The intent is to encode a BackupInfo object as an S3 object key. The
        goal of this is to retrieve all backups in a bucket with a single
        ListObjectsV2 call, which only provides keys, not any other metadata
        such as taggings.

        To construct such a key, callers should choose an arbitrary base name,
        and append the suffixes returned by this function, as well as any other
        suffixes as appropriate (such as .gz if the backup is gzipped).

        The order of suffixes is not significant for decoding, but they must
        all be present. The suffixes are returned in an order intended to make
        lists of backups more human-readable (this just means that the ctime is
        formatted as an ISO 8601 timestamp, and appears as the first suffix in
        the return value).

        The prefixes will start with '.'. It's suitable to join them into a
        filename like: f"basename{''.join(suffixes)}.gz".

        As the intent is to construct S3 keys, the return values are "regular"
        unicode strings, not fsdecode()'ed strings.

        Args:
            tzinfo: A time zone for formatting the ctime of the backup. The
                ctime will be output as an ISO 8601 string which always
                includes the time zone, so any value may be specified and the
                round-trip decoding won't be affected. For human readability,
                it's best to use the same time zone used for defining our
                preservation schedule.

        Returns:
            A sequence of suffix strings.
        """
        uuid = UUID(bytes=self.uuid)
        parent_uuid = UUID(bytes=self.parent_uuid)
        send_parent_uuid = (
            UUID(bytes=self.send_parent_uuid) if self.send_parent_uuid else UUID(int=0)
        )
        # arrow.get(float, tzinfo=None) raises an error, so we explicitly
        # default to UTC
        ctime = arrow.get(self.ctime, tzinfo="UTC" if tzinfo is None else tzinfo)
        return [
            f".{_CTIM}{ctime.isoformat(timespec='seconds')}",
            f".{_CTID}{self.ctransid}",
            f".{_UUID}{uuid}",
            f".{_SNDP}{send_parent_uuid}",
            f".{_PRNT}{parent_uuid}",
            f".{_MDVN}{_METADATA_VERSION}",
            f".{_SEQN}0",
        ]

    @classmethod
    def from_path(cls, path: str) -> Self:  # noqa: C901
        """Creates a BackupInfo from a backup filename or path.

        The path is expected to have a base name, and suffixes as returned by
        get_path_suffixes(). Other suffixes may be present, and will be
        ignored.

        The path is expected to be a "regular" unicode string, not a
        fsdecode()'ed string.

        Args:
            path: A filename or path for the backup.

        Returns:
            A BackupInfo object which had been previously encoded into an S3
            key.

        Raises:
            ValueError: If the input isn't a valid encoded BackupInfo.
        """
        uuid: UUID | None = None
        parent_uuid: UUID | None = None
        send_parent_uuid: UUID | None = None
        ctransid: int | None = None
        ctime: Arrow | None = None
        version: int | None = None
        sequence_number: int | None = None

        suffixes = pathlib.PurePath(path).suffixes
        for suffix in suffixes:
            code, rest = suffix[1:5], suffix[5:]
            if code == _PRNT:
                with suppress(ValueError):
                    parent_uuid = UUID(rest)
            elif code == _CTIM:
                with suppress(arrow.ParserError):
                    ctime = arrow.get(rest)
            elif code == _CTID:
                with suppress(ValueError):
                    ctransid = int(rest)
            elif code == _UUID:
                with suppress(ValueError):
                    uuid = UUID(rest)
            elif code == _SNDP:
                with suppress(ValueError):
                    send_parent_uuid = UUID(rest)
            elif code == _MDVN:
                with suppress(ValueError):
                    version = int(rest)
            elif code == _SEQN:
                with suppress(ValueError):
                    sequence_number = int(rest)

        if version is None:
            msg = "backup name metadata version missing (not a backup?)"
            raise ValueError(msg)
        if version != _METADATA_VERSION:
            msg = "unsupported backup name metadata version"
            raise ValueError(msg)
        if sequence_number != 0:
            msg = "unsupported sequence number"
            raise ValueError(msg)
        if (
            uuid is None
            or parent_uuid is None
            or ctransid is None
            or ctime is None
            or send_parent_uuid is None
        ):
            msg = "missing or incomplete parameters for backup name"
            raise ValueError(msg)
        return cls(
            uuid=uuid.bytes,
            parent_uuid=parent_uuid.bytes,
            ctransid=ctransid,
            ctime=ctime.timestamp(),
            send_parent_uuid=send_parent_uuid.bytes if send_parent_uuid.int else None,
        )



from __future__ import annotations

from typing import Protocol
from typing import TYPE_CHECKING
from uuid import uuid4

import arrow
from btrfs2s3._internal.preservation import Policy
from btrfs2s3._internal.resolver import _Resolver
from btrfs2s3._internal.resolver import Flags
from btrfs2s3._internal.resolver import KeepBackup
from btrfs2s3._internal.resolver import KeepMeta
from btrfs2s3._internal.resolver import Reasons
from btrfs2s3._internal.resolver import Result
from btrfs2s3._internal.util import backup_of_snapshot
from btrfs2s3._internal.util import mksubvol
import pytest

if TYPE_CHECKING:
    from btrfsutil import SubvolumeInfo


@pytest.fixture()
def parent_uuid() -> bytes:
    return uuid4().bytes


class MkSnap(Protocol):
    def __call__(self, *, t: str | None = None, i: int = 0) -> SubvolumeInfo: ...


@pytest.fixture()
def mksnap(parent_uuid: bytes) -> MkSnap:
    def inner(*, t: str | None = None, i: int = 0) -> SubvolumeInfo:
        a = arrow.get() if t is None else arrow.get(t)
        return mksubvol(
            uuid=uuid4().bytes, parent_uuid=parent_uuid, ctime=a.timestamp(), ctransid=i
        )

    return inner


def test_noop() -> None:
    resolver = _Resolver(snapshots=(), backups=(), policy=Policy())

    resolver.keep_send_ancestors_of_backups()

    assert resolver.get_result() == Result(keep_snapshots={}, keep_backups={})


def test_backup_with_no_parent(mksnap: MkSnap) -> None:
    snapshot1 = mksnap()
    backup1 = backup_of_snapshot(snapshot1, send_parent=None)
    resolver = _Resolver(snapshots=(), backups=(backup1,), policy=Policy())
    with resolver._keep_backups.with_reasons(Reasons.Preserved):
        resolver._keep_backups.mark(backup1)

    resolver.keep_send_ancestors_of_backups()

    assert resolver.get_result() == Result(
        keep_snapshots={},
        keep_backups={
            backup1.uuid: KeepBackup(backup1, KeepMeta(reasons=Reasons.Preserved))
        },
    )


def test_send_ancestors_already_kept(mksnap: MkSnap) -> None:
    snapshot1 = mksnap()
    snapshot2 = mksnap()
    snapshot3 = mksnap()
    backup1 = backup_of_snapshot(snapshot1, send_parent=None)
    backup2 = backup_of_snapshot(snapshot2, send_parent=snapshot1)
    backup3 = backup_of_snapshot(snapshot3, send_parent=snapshot2)
    resolver = _Resolver(
        snapshots=(), backups=(backup1, backup2, backup3), policy=Policy()
    )
    with resolver._keep_backups.with_reasons(Reasons.Preserved):
        resolver._keep_backups.mark(backup1)
        resolver._keep_backups.mark(backup2)
        resolver._keep_backups.mark(backup3)

    resolver.keep_send_ancestors_of_backups()

    assert resolver.get_result() == Result(
        keep_snapshots={},
        keep_backups={
            backup1.uuid: KeepBackup(backup1, KeepMeta(reasons=Reasons.Preserved)),
            backup2.uuid: KeepBackup(backup2, KeepMeta(reasons=Reasons.Preserved)),
            backup3.uuid: KeepBackup(backup3, KeepMeta(reasons=Reasons.Preserved)),
        },
    )


def test_send_ancestors_created_but_not_yet_kept(mksnap: MkSnap) -> None:
    snapshot1 = mksnap()
    snapshot2 = mksnap()
    snapshot3 = mksnap()
    backup1 = backup_of_snapshot(snapshot1, send_parent=None)
    backup2 = backup_of_snapshot(snapshot2, send_parent=snapshot1)
    backup3 = backup_of_snapshot(snapshot3, send_parent=snapshot2)
    resolver = _Resolver(
        snapshots=(snapshot1, snapshot2, snapshot3),
        backups=(backup1, backup2, backup3),
        policy=Policy(),
    )
    with resolver._keep_backups.with_reasons(Reasons.Preserved):
        resolver._keep_backups.mark(backup3)

    resolver.keep_send_ancestors_of_backups()

    assert resolver.get_result() == Result(
        keep_snapshots={},
        keep_backups={
            backup1.uuid: KeepBackup(
                backup1,
                KeepMeta(reasons=Reasons.SendAncestor, other_uuids={backup2.uuid}),
            ),
            backup2.uuid: KeepBackup(
                backup2,
                KeepMeta(reasons=Reasons.SendAncestor, other_uuids={backup3.uuid}),
            ),
            backup3.uuid: KeepBackup(backup3, KeepMeta(reasons=Reasons.Preserved)),
        },
    )


def test_send_ancestors_not_yet_created(mksnap: MkSnap) -> None:
    snapshot1 = mksnap()
    snapshot2 = mksnap()
    snapshot3 = mksnap()
    backup3 = backup_of_snapshot(snapshot3, send_parent=snapshot2)
    resolver = _Resolver(
        snapshots=(snapshot1, snapshot2, snapshot3),
        backups=(backup3,),
        policy=Policy.all(),
    )
    with resolver._keep_backups.with_reasons(Reasons.Preserved):
        resolver._keep_backups.mark(backup3)

    resolver.keep_send_ancestors_of_backups()

    expected_backup1 = backup_of_snapshot(snapshot1, send_parent=None)
    expected_backup2 = backup_of_snapshot(snapshot2, send_parent=snapshot1)
    assert resolver.get_result() == Result(
        keep_snapshots={},
        keep_backups={
            expected_backup1.uuid: KeepBackup(
                expected_backup1,
                KeepMeta(
                    reasons=Reasons.SendAncestor,
                    flags=Flags.New,
                    other_uuids={expected_backup2.uuid},
                ),
            ),
            expected_backup2.uuid: KeepBackup(
                expected_backup2,
                KeepMeta(
                    reasons=Reasons.SendAncestor,
                    flags=Flags.New,
                    other_uuids={backup3.uuid},
                ),
            ),
            backup3.uuid: KeepBackup(backup3, KeepMeta(reasons=Reasons.Preserved)),
        },
    )


def test_backup_chain_broken(mksnap: MkSnap) -> None:
    snapshot1 = mksnap()
    snapshot2 = mksnap()
    backup2 = backup_of_snapshot(snapshot2, send_parent=snapshot1)
    resolver = _Resolver(snapshots=(), backups=(backup2,), policy=Policy())
    with resolver._keep_backups.with_reasons(Reasons.Preserved):
        resolver._keep_backups.mark(backup2)

    with pytest.warns(UserWarning, match="Backup chain is broken"):
        resolver.keep_send_ancestors_of_backups()

    assert resolver.get_result() == Result(
        keep_snapshots={},
        keep_backups={
            backup2.uuid: KeepBackup(backup2, KeepMeta(reasons=Reasons.Preserved))
        },
    )



from __future__ import annotations

from btrfs2s3._internal.thunk import StillTBDError
from btrfs2s3._internal.thunk import TBD
from btrfs2s3._internal.thunk import Thunk
import pytest
from typing_extensions import assert_type


def test_immediate() -> None:
    t1 = Thunk("a")
    assert not t1.is_tbd()
    peek1 = t1.peek()
    assert peek1 is not TBD
    assert_type(peek1, str)
    assert peek1 == "a"
    assert t1.check() == "a"
    assert_type(t1(), str)
    assert t1() == "a"

    t2 = Thunk(t1)
    assert not t2.is_tbd()
    peek2 = t2.peek()
    assert peek2 is not TBD
    assert_type(peek2, str)
    assert peek2 == "a"
    assert t2.check() == "a"
    assert_type(t2(), str)
    assert t2() == "a"


def test_lazy() -> None:
    t1 = Thunk(lambda: "a")
    assert t1.is_tbd()
    peek1 = t1.peek()
    assert peek1 is TBD
    with pytest.raises(StillTBDError):
        t1.check()

    t2 = Thunk(t1)
    assert t1() == "a"
    assert_type(t1(), str)
    peek1 = t1.peek()
    assert not t1.is_tbd()
    assert peek1 is not TBD
    assert_type(peek1, str)
    assert peek1 == "a"

    assert t2.is_tbd()
    peek2 = t2.peek()
    assert peek2 is TBD
    with pytest.raises(StillTBDError):
        t2.check()
    assert t2() == "a"
    assert_type(t2(), str)


def test_repr() -> None:
    t_a = Thunk("a")
    assert repr(t_a) == "Thunk('a')"
    t_b = Thunk(lambda: "b")
    assert repr(t_b) == "Thunk(TBD)"
    t_b()
    assert repr(t_b) == "Thunk('b')"


def test_lt() -> None:
    t_a = Thunk("a")
    t_b = Thunk("b")
    assert t_a < t_b
    assert t_b > t_a

    t_tbd0 = Thunk(lambda: 0)
    t_tbd1 = Thunk(lambda: 1)
    assert (t_tbd0 < t_tbd1) != (t_tbd1 < t_tbd0)
    t_tbd1()
    assert t_tbd1 < t_tbd0
    assert t_tbd0 > t_tbd1
    t_tbd0()
    assert t_tbd0 < t_tbd1


def test_order() -> None:
    t_a = Thunk("a")
    t_b = Thunk("b")
    t_tbd = Thunk(lambda: "0")

    # When all values are the same type, they should obey their natural order
    assert sorted([t_b, t_a]) == [t_a, t_b]
    # When some thunks are TBD, they should appear at the end
    assert sorted([t_tbd, t_b, t_a]) == [t_a, t_b, t_tbd]
    t_tbd()
    assert sorted([t_tbd, t_b, t_a]) == [t_tbd, t_a, t_b]



"""Compatibility module for zoneinfo."""

import sys

if sys.version_info >= (3, 9):  # pragma: >=3.9 cover
    from zoneinfo import ZoneInfo
else:  # pragma: <3.9 cover
    from backports.zoneinfo import ZoneInfo


# This function is just syntactic sugar to call the ZoneInfo constructor with
# appropriate lint markers
def get_zoneinfo(key: str) -> ZoneInfo:
    """Return a ZoneInfo object from a string timezone name."""
    if sys.version_info >= (3, 9):  # pragma: >=3.9 cover
        return ZoneInfo(key)
    else:  # noqa: RET505 # pragma: <3.9 cover
        # https://github.com/pganssle/zoneinfo/issues/125
        # Unfortunately the workaround in the linked issue does not work with
        # mypy --strict (I get: Call to untyped function "ZoneInfo" in typed
        # context)

        # Also mypy requires if-return-else to understand the version
        # condition, which thus requires the RET505 block. Yakkity yak
        return ZoneInfo(key)  # type: ignore[abstract]



from collections.abc import Iterator
import os
import sys
from types import TracebackType
from typing import Any
from typing import Final
from typing import Literal
from typing import overload
from typing import TypeVar

from _typeshed import structseq
from typing_extensions import final
from typing_extensions import Self
from typing_extensions import TypeAlias

@final
class SubvolumeInfo(  # noqa: SLOT001
    structseq[Any],
    tuple[
        int, int, int, int, bytes, bytes, bytes, int, int, int, int, int, float, float
    ],
):
    if sys.version_info >= (3, 10):
        __match_args__: Final = (
            "id",
            "parent_id",
            "dir_id",
            "flags",
            "uuid",
            "parent_uuid",
            "received_uuid",
            "generation",
            "ctransid",
            "otransid",
            "stransid",
            "rtransid",
            "ctime",
            "otime",
        )
    @property
    def id(self) -> int: ...
    @property
    def parent_id(self) -> int: ...
    @property
    def dir_id(self) -> int: ...
    @property
    def flags(self) -> int: ...
    @property
    def uuid(self) -> bytes: ...
    @property
    def parent_uuid(self) -> bytes: ...
    @property
    def received_uuid(self) -> bytes: ...
    @property
    def generation(self) -> int: ...
    @property
    def ctransid(self) -> int: ...
    @property
    def otransid(self) -> int: ...
    @property
    def stransid(self) -> int: ...
    @property
    def rtransid(self) -> int: ...
    @property
    def ctime(self) -> float: ...
    @property
    def otime(self) -> float: ...
    @property
    def stime(self) -> float: ...
    @property
    def rtime(self) -> float: ...

_IT_co = TypeVar("_IT_co", covariant=True, bound=int | SubvolumeInfo)
_Path: TypeAlias = str | bytes | os.PathLike[str] | os.PathLike[bytes]

class SubvolumeIterator(Iterator[tuple[str, _IT_co]]):
    @overload
    def __new__(
        cls,
        path: _Path | int,
        *,
        info: Literal[True],
        top: int = 0,
        post_order: bool = False,
    ) -> SubvolumeIterator[SubvolumeInfo]: ...
    @overload
    def __new__(
        cls,
        path: _Path | int,
        *,
        info: Literal[False],
        top: int = 0,
        post_order: bool = False,
    ) -> SubvolumeIterator[int]: ...
    @overload
    def __new__(
        cls, path: _Path | int, *, top: int = 0, post_order: bool = False
    ) -> SubvolumeIterator[int]: ...
    def __next__(self) -> tuple[str, _IT_co]: ...
    def __iter__(self) -> Self: ...
    def close(self) -> None: ...
    def fileno(self) -> int: ...
    def __enter__(self) -> Self: ...
    def __exit__(
        self,
        __exc_type: type[BaseException] | None,
        __exc_value: BaseException | None,
        __traceback: TracebackType | None,
    ) -> bool | None: ...

class QgroupInherit:
    @property
    def groups(self) -> list[int]: ...
    def add_group(self, qgroupid: int) -> None: ...

@overload
def create_snapshot(
    source: _Path | int,
    path: _Path,
    *,
    async_: Literal[True],
    recursive: bool = False,
    read_only: bool = False,
    qgroup_inherit: QgroupInherit | None = None,
) -> int: ...
@overload
def create_snapshot(
    source: _Path | int,
    path: _Path,
    *,
    async_: Literal[False],
    recursive: bool = False,
    read_only: bool = False,
    qgroup_inherit: QgroupInherit | None = None,
) -> None: ...
@overload
def create_snapshot(
    source: _Path | int,
    path: _Path,
    *,
    recursive: bool = False,
    read_only: bool = False,
    qgroup_inherit: QgroupInherit | None = None,
) -> None: ...
@overload
def create_subvolume(
    path: _Path, *, async_: Literal[True], qgroup_inherit: QgroupInherit | None = None
) -> int: ...
@overload
def create_subvolume(
    path: _Path, *, async_: Literal[False], qgroup_inherit: QgroupInherit | None = None
) -> None: ...
@overload
def create_subvolume(
    path: _Path, *, qgroup_inherit: QgroupInherit | None = None
) -> None: ...
def is_subvolume(path: _Path | int) -> bool: ...
def subvolume_info(path: _Path | int, id: int = 0) -> SubvolumeInfo: ...
def delete_subvolume(path: _Path, *, recursive: bool = False) -> None: ...
def sync(path: _Path | int) -> None: ...
def set_subvolume_read_only(path: _Path | int, read_only: bool) -> None: ...  # noqa: FBT001



"""Code for manipulating config for btrfs2s3."""

from __future__ import annotations

from collections import namedtuple
from typing import Any
from typing import cast
from typing import TYPE_CHECKING
from typing import TypedDict

from cfgv import Array
from cfgv import check_array
from cfgv import check_string
from cfgv import check_type
from cfgv import load_from_filename
from cfgv import Map
from cfgv import OptionalNoDefault
from cfgv import OptionalRecurse
from cfgv import Required
from cfgv import RequiredRecurse
from typing_extensions import NotRequired
from yaml import safe_load

from btrfs2s3._internal.preservation import Params

if TYPE_CHECKING:
    from os import PathLike


class Error(Exception):
    """The top-level class for exceptions generated by this module."""


class InvalidConfigError(Error):
    """An error for invalid config."""


def _check_preserve(v: Any) -> None:  # noqa: ANN401
    check_string(v)
    try:
        Params.parse(v)
    except ValueError as ex:
        msg = "Expected a valid schedule"
        raise InvalidConfigError(msg) from ex


# this is the same style used in cfgv
_OptionalRecurseNoDefault = namedtuple(  # noqa: PYI024
    "_OptionalRecurseNoDefault", ("key", "schema")
)
_OptionalRecurseNoDefault.check = OptionalRecurse.check  # type: ignore[attr-defined]
_OptionalRecurseNoDefault.check_fn = OptionalRecurse.check_fn  # type: ignore[attr-defined]
_OptionalRecurseNoDefault.apply_default = OptionalNoDefault.apply_default  # type: ignore[attr-defined]
_OptionalRecurseNoDefault.remove_default = OptionalNoDefault.remove_default  # type: ignore[attr-defined]


class S3EndpointConfig(TypedDict):
    """A config dict for how to talk to an S3 endpoint."""

    aws_access_key_id: NotRequired[str]
    aws_secret_access_key: NotRequired[str]
    region_name: NotRequired[str]
    profile_name: NotRequired[str]
    verify: NotRequired[bool | str]
    endpoint_url: NotRequired[str]


_S3_ENDPOINT_SCHEMA = Map(
    "S3EndpointConfig",
    None,
    OptionalNoDefault("aws_access_key_id", check_string),
    OptionalNoDefault("aws_secret_access_key", check_string),
    OptionalNoDefault("region_name", check_string),
    OptionalNoDefault("profile_name", check_string),
    OptionalNoDefault("verify", check_type((bool, str), typename="bool or path")),
    OptionalNoDefault("endpoint_url", check_string),
)


class S3RemoteConfig(TypedDict):
    """A config dict for how to access an S3 remote."""

    bucket: str
    endpoint: NotRequired[S3EndpointConfig]


_S3_SCHEMA = Map(
    "S3RemoteConfig",
    None,
    Required("bucket", check_string),
    _OptionalRecurseNoDefault("endpoint", _S3_ENDPOINT_SCHEMA),
)


class RemoteConfig(TypedDict):
    """A config dict for how to access a remote."""

    id: str
    s3: NotRequired[S3RemoteConfig]


_REMOTE_SCHEMA = Map(
    "RemoteConfig",
    None,
    Required("id", check_string),
    RequiredRecurse("s3", _S3_SCHEMA),
)


class UploadToRemoteConfig(TypedDict):
    """A config dict for uploading a source to a remote."""

    id: str
    preserve: str
    pipe_through: NotRequired[list[list[str]]]


_UPLOAD_TO_REMOTE_SCHEMA = Map(
    "UploadToRemoteConfig",
    "id",
    Required("preserve", _check_preserve),
    OptionalNoDefault("pipe_through", check_array(check_array(check_string))),
)


class SourceConfig(TypedDict):
    """A config dict for a source."""

    path: str
    snapshots: str
    upload_to_remotes: list[UploadToRemoteConfig]


_SOURCE_SCHEMA = Map(
    "SourceConfig",
    "path",
    Required("path", check_string),
    Required("snapshots", check_string),
    RequiredRecurse(
        "upload_to_remotes", Array(_UPLOAD_TO_REMOTE_SCHEMA, allow_empty=False)
    ),
)


class Config(TypedDict):
    """The top-level config dict.

    This just matches the data as it's stored in config.yaml. We don't do any
    transformation up front (for example "preserve" values are just their
    strings, not Policy objects).
    """

    timezone: str
    sources: list[SourceConfig]
    remotes: list[RemoteConfig]


_SCHEMA = Map(
    "Config",
    None,
    Required("timezone", check_string),
    RequiredRecurse("sources", Array(_SOURCE_SCHEMA, allow_empty=False)),
    RequiredRecurse("remotes", Array(_REMOTE_SCHEMA, allow_empty=False)),
)


def load_from_path(path: str | PathLike[str]) -> Config:
    """Load config from a file path.

    This performs some basic syntactic validation on the config, to ensure it
    really conforms to the return type.

    Args:
        path: The path to the config.

    Returns:
        A Config instance.

    Raises:
        InvalidConfigError: If the config does not pass validation.
    """
    config = cast(
        Config, load_from_filename(path, _SCHEMA, safe_load, exc_tp=InvalidConfigError)
    )

    remote_ids = {remote["id"] for remote in config["remotes"]}
    for source in config["sources"]:
        for upload_to_remote in source["upload_to_remotes"]:
            if upload_to_remote["id"] not in remote_ids:
                msg = (
                    f'remote id {upload_to_remote["id"]!r} for source '
                    f'{source["path"]!r} is not defined in the list of remotes'
                )
                raise InvalidConfigError(msg)

    # https://github.com/sbrudenell/btrfs2s3/issues/29
    if len(config["remotes"]) > 1:
        msg = "multiple remotes not supported in this release"
        raise InvalidConfigError(msg)

    all_sources = config["sources"]

    # https://github.com/sbrudenell/btrfs2s3/issues/81
    if len({source["snapshots"] for source in all_sources}) > 1:
        msg = "multiple snapshot locations not supported in this release"
        raise InvalidConfigError(msg)

    all_uploads = [up for src in all_sources for up in src["upload_to_remotes"]]

    # https://github.com/sbrudenell/btrfs2s3/issues/79
    if len({upload["preserve"] for upload in all_uploads}) > 1:
        msg = "multiple preserve configurations not supported in this release"
        raise InvalidConfigError(msg)

    all_pipe_throughs = [up.get("pipe_through", []) for up in all_uploads]

    # https://github.com/sbrudenell/btrfs2s3/issues/80
    if len({tuple(tuple(cmd) for cmd in p) for p in all_pipe_throughs}) > 1:
        msg = "multiple pipe_through configurations not supported in this release"
        raise InvalidConfigError(msg)

    return config



"""Code for "btrfs2s3 update"."""

from __future__ import annotations

from collections import defaultdict
from pathlib import Path
from typing import cast
from typing import TYPE_CHECKING

import arrow
from boto3.session import Session
from rich.box import HORIZONTALS
from rich.console import Console
from rich.console import Group
from rich.highlighter import ISO8601Highlighter
from rich.panel import Panel
from rich.prompt import Confirm
from rich.rule import Rule
from rich.table import Table
from rich.text import Text
from rich.tree import Tree

from btrfs2s3._internal.action import Actions
from btrfs2s3._internal.assessor import assess
from btrfs2s3._internal.assessor import Assessment
from btrfs2s3._internal.assessor import assessment_to_actions
from btrfs2s3._internal.assessor import BackupAssessment
from btrfs2s3._internal.assessor import SourceAssessment
from btrfs2s3._internal.config import Config
from btrfs2s3._internal.config import load_from_path
from btrfs2s3._internal.preservation import Params
from btrfs2s3._internal.preservation import Policy
from btrfs2s3._internal.preservation import TS
from btrfs2s3._internal.resolver import Flags
from btrfs2s3._internal.resolver import KeepMeta
from btrfs2s3._internal.resolver import Reasons
from btrfs2s3._internal.thunk import TBD
from btrfs2s3._internal.zoneinfo import get_zoneinfo

if TYPE_CHECKING:
    import argparse
    from datetime import tzinfo
    from typing import Iterable
    from typing import Literal
    from typing import Sequence
    from typing import TypedDict

    from typing_extensions import TypeAlias

    _Bounds: TypeAlias = Literal["[)", "()", "(]", "[]"]

_iso8601_highlight = ISO8601Highlighter()


def _time_span_key(time_span: TS) -> tuple[float, float]:
    start, end = time_span
    return (start - end, start)


def _describe_time_spans(time_spans: Iterable[TS], tzinfo: tzinfo) -> Text:
    return describe_time_span(
        sorted(time_spans, key=_time_span_key)[0], tzinfo, bounds="[]"
    )


def describe_time_span(
    time_span: TS, tzinfo: tzinfo, *, bounds: _Bounds = "[)"
) -> Text:
    """Returns a highlighted summary of a time span in context of preservation."""
    a_timestamp, b_timestamp = time_span
    a = arrow.get(a_timestamp, tzinfo=tzinfo)
    b = arrow.get(b_timestamp, tzinfo=tzinfo)
    if (a, b) == a.span("year", bounds=bounds):
        return Text.from_markup(f"[iso8601.date]{a.year:04d}[/] yearly")
    if (a, b) == a.span("quarter", bounds=bounds):
        return Text.from_markup(f"[iso8601.date]{a.year:04d}-Q{a.quarter}[/] quarterly")
    if (a, b) == a.span("month", bounds=bounds):
        return Text.from_markup(f"[iso8601.date]{a.year:04d}-{a.month:02d}[/] monthly")
    if (a, b) == a.span("week", bounds=bounds):
        return Text.from_markup(f"[iso8601.date]{a.year:04d}-W{a.week:02d}[/] weekly")
    if (a, b) == a.span("day", bounds=bounds):
        return Text.from_markup(
            f"[iso8601.date]{a.year:04d}-{a.month:02d}-{a.day:02d}[/] daily"
        )
    if (a, b) == a.span("hour", bounds=bounds):
        return Text.from_markup(
            f"[iso8601.date]{a.year:04d}-{a.month:02d}-{a.day:02d}[/]T"
            f"[iso8601.time]{a.hour:02d}[/] hourly"
        )
    if (a, b) == a.span("minute", bounds=bounds):
        return Text.from_markup(
            f"[iso8601.date]{a.year:04d}-{a.month:02d}-{a.day:02d}[/]T"
            f"[iso8601.time]{a.hour:02d}:{a.minute:02d}[/] minutely"
        )
    if (a, b) == a.span("second", bounds=bounds):
        return Text.from_markup(
            f"[iso8601.date]{a.year:04d}-{a.month:02d}-{a.day:02d}[/]T"
            f"[iso8601.time]{a.hour:02d}:{a.minute:02d}:{a.second:02d}[/] secondly"
        )
    return (
        _iso8601_highlight(a.format("YYYY-MM-DDTHH:mm:ss"))
        .append("/")
        .append(_iso8601_highlight(b.format("YYYY-MM-DDTHH:mm:ss")))
    )


def _describe_preserve(keep_meta: KeepMeta, tzinfo: tzinfo) -> Text:
    if keep_meta.reasons & Reasons.Preserved:
        return _describe_time_spans(keep_meta.time_spans, tzinfo)
    if keep_meta.reasons & Reasons.MostRecent:
        return Text("<most recent>", style="keep")
    if keep_meta.reasons & Reasons.SendAncestor:
        return Text("<ancestor>", style="keep")
    if keep_meta.reasons:  # pragma: no cover
        return Text("<keep!>", style="keep")
    return Text("<not keeping>", style="not_keeping")


def _describe_time(time: float, tzinfo: tzinfo) -> Text:
    return _iso8601_highlight(
        arrow.get(time, tzinfo=tzinfo).format("YYYY-MM-DDTHH:mm:ss")
    )


def _keep_emoji(keep_meta: KeepMeta) -> Text:
    if keep_meta.flags & Flags.New:
        return Text.from_markup(":sparkles:")
    if not keep_meta.reasons:
        return Text.from_markup(":skull:")
    return Text()


def _backup_asmt_key(asmt: BackupAssessment) -> tuple[bool, int]:
    info = asmt.backup.peek()
    return (asmt.new, 0 if info is TBD else info.ctransid)


def _make_snapshots_table(
    *, asmt: SourceAssessment, tzinfo: tzinfo, snapshot_dir: Path
) -> Table:
    table = Table(
        title=f"snapshots of {asmt.path} in {snapshot_dir}",
        title_justify="left",
        row_styles=["none", "dim"],
        box=HORIZONTALS,
    )
    table.add_column("path", no_wrap=True)
    table.add_column("ctime", no_wrap=True)
    table.add_column("ctransid", style="ctransid", no_wrap=True)
    table.add_column("preserve", no_wrap=True)
    for snapshot in sorted(asmt.snapshots.values(), key=lambda a: a.info.ctransid):
        path = _keep_emoji(snapshot.keep_meta).append(str(snapshot.initial_path))
        ctime = _describe_time(snapshot.info.ctime, tzinfo)
        ctransid = str(snapshot.info.ctransid)
        preserve = _describe_preserve(snapshot.keep_meta, tzinfo)
        table.add_row(path, ctime, ctransid, preserve)
    return table


def _make_backups_tree(*, asmt: SourceAssessment, tzinfo: tzinfo, bucket: str) -> Tree:
    uuid_to_children: dict[bytes | None, list[bytes]] = defaultdict(list)
    for uuid, backup in asmt.backups.items():
        parent: bytes | None = None
        if not backup.backup.is_tbd():
            parent = backup.backup().send_parent_uuid
        uuid_to_children[parent].append(uuid)

    def add_children(tree: Tree, parent: bytes | None) -> None:
        children = {u: asmt.backups[u] for u in uuid_to_children.get(parent, ())}
        for uuid, backup in sorted(
            children.items(), key=lambda i: _backup_asmt_key(i[1])
        ):
            lines = []

            key = Text("key: ", no_wrap=True)
            key.append(_keep_emoji(backup.keep_meta))
            key.append(str(backup.key.peek()), style="key")

            if not backup.backup.is_tbd():
                info = Text(no_wrap=True)
                info.append(_describe_time(backup.backup().ctime, tzinfo))
                info.append(" / ctransid ")
                info.append(str(backup.backup().ctransid), style="ctransid")
                lines.append(info)

            preserve = Text(no_wrap=True)
            preserve.append(_describe_preserve(backup.keep_meta, tzinfo))
            lines.append(preserve)

            if uuid in asmt.snapshots:
                origin = Text(no_wrap=True)
                origin.append("backup of: ")
                origin.append(_keep_emoji(asmt.snapshots[uuid].keep_meta))
                origin.append(str(asmt.snapshots[uuid].initial_path))
                lines.append(origin)

            subtree = tree.add(Panel(Group(*lines), title=key, title_align="left"))
            add_children(subtree, uuid)

    tree = Tree(f"tree of incremental backups of {asmt.path} in bucket {bucket}")
    add_children(tree, None)
    return tree


def _print_source_assessment(
    *,
    console: Console,
    asmt: SourceAssessment,
    tzinfo: tzinfo,
    snapshot_dir: Path,
    bucket: str,
) -> None:
    console.print(
        _make_snapshots_table(asmt=asmt, tzinfo=tzinfo, snapshot_dir=snapshot_dir)
    )
    console.print()
    console.print(_make_backups_tree(asmt=asmt, tzinfo=tzinfo, bucket=bucket))
    console.print()


def print_assessment(
    *,
    console: Console,
    asmt: Assessment,
    tzinfo: tzinfo,
    snapshot_dir: Path,
    bucket: str,
) -> None:
    """Prints an Assessment to a Console."""
    console.print(Rule("assessment and proposed new state"))
    console.print()
    for source_asmt in sorted(asmt.sources.values(), key=lambda a: a.path):
        _print_source_assessment(
            console=console,
            asmt=source_asmt,
            tzinfo=tzinfo,
            snapshot_dir=snapshot_dir,
            bucket=bucket,
        )
        console.print()


def _make_action_tables(actions: Actions) -> Sequence[Table]:
    create_snapshots = Table(
        "action",
        "source",
        "arrow",
        "path",
        box=None,
        show_header=False,
        row_styles=["none", "dim"],
    )
    for csi in actions.iter_create_snapshot_intents():
        create_snapshots.add_row(
            Text.from_markup(":camera_with_flash: create snapshot:"),
            str(csi.source.peek()),
            "->",
            str(csi.path.peek()),
            style="create",
        )

    rename_snapshots = Table(
        "action",
        "source",
        "arrow",
        "dest",
        box=None,
        show_header=False,
        row_styles=["none", "dim"],
    )
    for rsi in actions.iter_rename_snapshot_intents():
        rename_snapshots.add_row(
            Text.from_markup(":pencil-emoji: rename snapshot:"),
            str(rsi.source.peek()),
            "->",
            str(rsi.target.peek()),
            style="modify",
        )

    create_backups = Table(
        "action", "source", box=None, show_header=False, row_styles=["none", "dim"]
    )
    for cbi in actions.iter_create_backup_intents():
        if cbi.snapshot.is_tbd():
            source = f"(new snapshot of {cbi.source.peek()})"
        else:
            source = str(cbi.snapshot.peek())
        create_backups.add_row(
            Text.from_markup(":cloud-emoji: upload backup:"), source, style="create"
        )

    delete_snapshots = Table(
        "action", "path", box=None, show_header=False, row_styles=["none", "dim"]
    )
    for dsi in actions.iter_delete_snapshot_intents():
        delete_snapshots.add_row(
            Text.from_markup(":skull: delete snapshot:"),
            str(dsi.path.peek()),
            style="delete",
        )

    delete_backups = Table(
        "action", "key", box=None, show_header=False, row_styles=["none", "dim"]
    )
    for dbi in actions.iter_delete_backup_intents():
        delete_backups.add_row(
            Text.from_markup(":skull: delete backup: "),
            str(dbi.key.peek()),
            style="delete",
        )

    return (
        create_snapshots,
        rename_snapshots,
        create_backups,
        delete_snapshots,
        delete_backups,
    )


def print_actions(*, console: Console, actions: Actions) -> None:
    """Prints a list of Actions to a Console."""
    if actions.empty():
        return

    console.print(Rule("actions to be taken"))

    tables = _make_action_tables(actions)

    for table in tables:
        if table.row_count:
            console.print(table)
            console.print()


NAME = "update"


if TYPE_CHECKING:

    class _Args(TypedDict, total=False):
        help: str
        description: str
        epilog: str


ARGS: _Args = {
    # shown in top-level help
    "help": "one-time update of snapshots and backups",
    # shown in subcommand help
    "description": "One-time update of snapshots and backups.",
    "epilog": "For detailed docs and usage, see https://github.com/sbrudenell/btrfs2s3",
}


def add_args(parser: argparse.ArgumentParser) -> None:
    """Add args for "btrfs2s3 update" to an ArgumentParser."""
    parser.add_argument("config_file", type=load_from_path)
    parser.add_argument(
        "--force", action="store_true", help="perform actions without prompting"
    )
    parser.add_argument(
        "--pretend",
        action="store_true",
        help="do not perform actions, just print a preview and exit",
    )


def command(*, console: Console, args: argparse.Namespace) -> int:
    """Implements "btrfs2s3 update"."""
    if not console.is_terminal and not (args.force or args.pretend):
        console.print("to run in unattended mode, use --force")
        return 1

    config = cast(Config, args.config_file)
    tzinfo = get_zoneinfo(config["timezone"])
    s3_remote = config["remotes"][0]["s3"]
    s3_endpoint = s3_remote.get("endpoint", {})

    sources = config["sources"]

    session = Session(
        region_name=s3_endpoint.get("region_name"),
        profile_name=s3_endpoint.get("profile_name"),
    )
    s3 = session.client(
        "s3",
        # https://docs.aws.amazon.com/sdk-for-javascript/v2/developer-guide/locking-api-versions.html
        # says that pinning api_version is a best practice
        api_version="2006-03-01",
        verify=s3_endpoint.get("verify"),
        endpoint_url=s3_endpoint.get("endpoint_url"),
    )
    policy = Policy(
        tzinfo=tzinfo,
        params=Params.parse(sources[0]["upload_to_remotes"][0]["preserve"]),
    )
    asmt = assess(
        snapshot_dir=Path(sources[0]["snapshots"]),
        sources=[Path(source["path"]) for source in sources],
        s3=s3,
        bucket=s3_remote["bucket"],
        policy=policy,
    )
    actions = Actions()
    assessment_to_actions(asmt, actions)

    if console.is_terminal:
        print_assessment(
            console=console,
            asmt=asmt,
            tzinfo=tzinfo,
            snapshot_dir=Path(sources[0]["snapshots"]),
            bucket=s3_remote["bucket"],
        )
        print_actions(console=console, actions=actions)

    if args.pretend:
        return 0

    if actions.empty():
        console.print("nothing to be done!")
        return 0

    if args.force or Confirm(console=console).ask("continue?"):
        actions.execute(
            s3,
            s3_remote["bucket"],
            pipe_through=sources[0]["upload_to_remotes"][0].get("pipe_through", []),
        )

    return 0



from __future__ import annotations

from base64 import b64decode
import gzip
import subprocess
from typing import TYPE_CHECKING

from botocore.exceptions import ClientError
from btrfs2s3._internal.action import create_backup
import btrfsutil
import pytest

if TYPE_CHECKING:
    from pathlib import Path

    from mypy_boto3_s3.client import S3Client

    from tests.conftest import DownloadAndPipe


def test_creates_a_valid_btrfs_archive(
    btrfs_mountpoint: Path,
    s3: S3Client,
    bucket: str,
    download_and_pipe: DownloadAndPipe,
) -> None:
    source = btrfs_mountpoint / "source"
    btrfsutil.create_subvolume(source)

    snapshot = btrfs_mountpoint / "snapshot"
    btrfsutil.create_snapshot(source, snapshot, read_only=True)

    key = "test-backup"

    create_backup(s3=s3, bucket=bucket, key=key, snapshot=snapshot, send_parent=None)

    # Just check the archive is valid
    download_and_pipe(key, ["btrfs", "receive", "--dump"])


def test_large_archive_multipart_upload(
    btrfs_mountpoint: Path,
    s3: S3Client,
    bucket: str,
    download_and_pipe: DownloadAndPipe,
) -> None:
    source = btrfs_mountpoint / "source"
    btrfsutil.create_subvolume(source)

    # The default multipart threshold is 8MB. So create a 16mb file
    (source / "large-file").write_bytes(b"\xff" * (16 * 2**20))

    snapshot = btrfs_mountpoint / "snapshot"
    btrfsutil.create_snapshot(source, snapshot, read_only=True)

    key = "test-backup"

    create_backup(s3=s3, bucket=bucket, snapshot=snapshot, send_parent=None, key=key)

    # Just check the archive is valid
    download_and_pipe(key, ["btrfs", "receive", "--dump"])


def test_send_full_and_delta_archives_and_restore(
    btrfs_mountpoint: Path,
    s3: S3Client,
    bucket: str,
    download_and_pipe: DownloadAndPipe,
) -> None:
    source = btrfs_mountpoint / "source"
    btrfsutil.create_subvolume(source)
    large_file = source / "large-file"

    # Write some data in the source
    large_file.write_bytes(b"\xff" * (16 * 2**20))
    # Create a snapshot
    snapshot1 = btrfs_mountpoint / "snapshot1"
    btrfsutil.create_snapshot(source, snapshot1, read_only=True)
    # Back up the first snapshot
    key1 = "test-backup1"
    create_backup(s3=s3, bucket=bucket, snapshot=snapshot1, send_parent=None, key=key1)

    # Write some more data in the source (appending this time)
    with large_file.open(mode="ab") as fp:
        fp.write(b"\xff" * (16 * 2**20))

    # Create another snapshot
    snapshot2 = btrfs_mountpoint / "snapshot2"
    btrfsutil.create_snapshot(source, snapshot2, read_only=True)
    # Back up the first snapshot, using the first as a send-parent
    key2 = "test-backup2"
    create_backup(
        s3=s3, bucket=bucket, snapshot=snapshot2, send_parent=snapshot1, key=key2
    )

    # Delete the snapshots and the source
    btrfsutil.delete_subvolume(source)
    btrfsutil.delete_subvolume(snapshot1)
    btrfsutil.delete_subvolume(snapshot2)

    # Restore the snapshots
    download_and_pipe(key1, ["btrfs", "receive", btrfs_mountpoint])
    download_and_pipe(key2, ["btrfs", "receive", btrfs_mountpoint])

    # The restored latest snapshot should contain the full set of data
    assert (snapshot2 / "large-file").read_bytes() == b"\xff" * (32 * 2**20)


def test_raise_error_on_send_failure(
    tmp_path: Path, capfd: pytest.CaptureFixture[str], s3: S3Client, bucket: str
) -> None:
    snapshot = tmp_path / "not-a-btrfs-subvolume"
    key = "test-backup"

    with pytest.raises(RuntimeError, match="exited with code "):
        create_backup(
            s3=s3, bucket=bucket, snapshot=snapshot, send_parent=None, key=key
        )
    out, err = capfd.readouterr()
    assert out == ""
    assert err != ""

    # Check delete
    with pytest.raises(ClientError):
        s3.head_object(Bucket=bucket, Key=key)


def test_pipe_success(btrfs_mountpoint: Path, s3: S3Client, bucket: str) -> None:
    source = btrfs_mountpoint / "source"
    btrfsutil.create_subvolume(source)

    snapshot = btrfs_mountpoint / "snapshot"
    btrfsutil.create_snapshot(source, snapshot, read_only=True)

    key = "test-backup"

    create_backup(
        s3=s3,
        bucket=bucket,
        key=key,
        snapshot=snapshot,
        send_parent=None,
        pipe_through=[["gzip"], ["base64"]],
    )

    # Just check the archive is valid
    compressed_encoded_data = s3.get_object(Bucket=bucket, Key=key)["Body"].read()
    data = gzip.decompress(b64decode(compressed_encoded_data))
    subprocess.run(["btrfs", "receive", "--dump"], input=data, check=True)


def test_end_of_pipe_fails(
    btrfs_mountpoint: Path, s3: S3Client, bucket: str, capfd: pytest.CaptureFixture[str]
) -> None:
    source = btrfs_mountpoint / "source"
    btrfsutil.create_subvolume(source)

    snapshot = btrfs_mountpoint / "snapshot"
    btrfsutil.create_snapshot(source, snapshot, read_only=True)

    key = "test-backup"

    with pytest.raises(RuntimeError, match="exited with code "):
        create_backup(
            s3=s3,
            bucket=bucket,
            snapshot=snapshot,
            send_parent=None,
            key=key,
            pipe_through=[["gzip"], ["base64", "--bad-option"]],
        )
    out, err = capfd.readouterr()
    assert out == ""
    assert err != ""

    # Check delete
    with pytest.raises(ClientError):
        s3.head_object(Bucket=bucket, Key=key)


def test_middle_of_pipe_fails(
    btrfs_mountpoint: Path, s3: S3Client, bucket: str, capfd: pytest.CaptureFixture[str]
) -> None:
    source = btrfs_mountpoint / "source"
    btrfsutil.create_subvolume(source)

    snapshot = btrfs_mountpoint / "snapshot"
    btrfsutil.create_snapshot(source, snapshot, read_only=True)

    key = "test-backup"

    with pytest.raises(RuntimeError, match="exited with code "):
        create_backup(
            s3=s3,
            bucket=bucket,
            snapshot=snapshot,
            send_parent=None,
            key=key,
            pipe_through=[["gzip", "--bad-option"], ["base64"]],
        )
    out, err = capfd.readouterr()
    assert out == ""
    assert err != ""

    # Check delete
    with pytest.raises(ClientError):
        s3.head_object(Bucket=bucket, Key=key)


def test_start_of_pipe_fails(
    tmp_path: Path, s3: S3Client, bucket: str, capfd: pytest.CaptureFixture[str]
) -> None:
    snapshot = tmp_path / "not-a-btrfs-subvolume"
    key = "test-backup"

    with pytest.raises(RuntimeError, match="exited with code "):
        create_backup(
            s3=s3,
            bucket=bucket,
            snapshot=snapshot,
            send_parent=None,
            key=key,
            pipe_through=[["gzip"], ["base64"]],
        )
    out, err = capfd.readouterr()
    assert out == ""
    assert err != ""

    # Check delete
    with pytest.raises(ClientError):
        s3.head_object(Bucket=bucket, Key=key)



from __future__ import annotations

from uuid import uuid4

from btrfs2s3._internal.backups import BackupInfo
from btrfs2s3._internal.preservation import Policy
from btrfs2s3._internal.resolver import _Index
from btrfs2s3._internal.util import mksubvol


def test_get_nominal_none() -> None:
    index = _Index(policy=Policy(), items=())
    got = index.get_nominal((0.0, 0.0))
    assert got is None


def test_get_nominal_snapshot_one() -> None:
    snapshot = mksubvol()
    index = _Index(policy=Policy.all(), items=(snapshot,))
    for timespan in Policy.all().iter_time_spans(snapshot.ctime):
        got = index.get_nominal(timespan)
        assert got == snapshot


def test_get_nominal_snapshot_one_of_many() -> None:
    snapshot1 = mksubvol(ctransid=1)
    snapshot2 = mksubvol(ctransid=2)
    index = _Index(policy=Policy.all(), items=(snapshot1, snapshot2))
    for timespan in Policy.all().iter_time_spans(snapshot1.ctime):
        got = index.get_nominal(timespan)
        assert got == snapshot1


def test_get_nominal_backup_one() -> None:
    backup = BackupInfo(
        uuid=uuid4().bytes,
        parent_uuid=uuid4().bytes,
        send_parent_uuid=None,
        ctime=0,
        ctransid=0,
    )
    index = _Index(policy=Policy.all(), items=(backup,))
    for timespan in Policy.all().iter_time_spans(backup.ctime):
        got = index.get_nominal(timespan)
        assert got == backup


def test_get_nominal_backup_one_of_many() -> None:
    backup1 = BackupInfo(
        uuid=uuid4().bytes,
        parent_uuid=uuid4().bytes,
        send_parent_uuid=None,
        ctime=0,
        ctransid=0,
    )
    backup2 = BackupInfo(
        uuid=uuid4().bytes,
        parent_uuid=backup1.parent_uuid,
        send_parent_uuid=None,
        ctime=0,
        ctransid=1,
    )
    index = _Index(policy=Policy.all(), items=(backup1, backup2))
    for timespan in Policy.all().iter_time_spans(backup1.ctime):
        got = index.get_nominal(timespan)
        assert got == backup1


def test_get_none() -> None:
    index = _Index(policy=Policy(), items=())
    got = index.get(uuid4().bytes)
    assert got is None


def test_get_snapshot() -> None:
    snapshot = mksubvol(uuid=uuid4().bytes)
    index = _Index(policy=Policy(), items=(snapshot,))
    got = index.get(snapshot.uuid)
    assert got == snapshot


def test_get_backup() -> None:
    backup = BackupInfo(
        uuid=uuid4().bytes,
        parent_uuid=uuid4().bytes,
        send_parent_uuid=None,
        ctime=0,
        ctransid=0,
    )
    index = _Index(policy=Policy(), items=(backup,))
    got = index.get(backup.uuid)
    assert got == backup


def test_get_most_recent_none() -> None:
    index = _Index(policy=Policy(), items=())
    got = index.get_most_recent()
    assert got is None


def test_get_most_recent_snapshot() -> None:
    snapshot1 = mksubvol(uuid=uuid4().bytes, ctransid=1)
    snapshot2 = mksubvol(uuid=uuid4().bytes, ctransid=2)
    index = _Index(policy=Policy(), items=(snapshot1, snapshot2))
    got = index.get_most_recent()
    assert got == snapshot2


def test_get_all_time_spans_empty() -> None:
    index = _Index(policy=Policy.all(), items=())
    got = index.get_all_time_spans()
    assert set(got) == set()


def test_get_all_time_spans() -> None:
    snapshot = mksubvol(uuid=uuid4().bytes, ctime=2000000000.0)
    index = _Index(policy=Policy.all(), items=(snapshot,))
    got = index.get_all_time_spans()
    assert got == set(Policy.all().iter_time_spans(snapshot.ctime))



from __future__ import annotations

from random import uniform
import time
from typing import cast

import arrow
from btrfs2s3._internal.arrowutil import convert_span
from btrfs2s3._internal.preservation import Params
from btrfs2s3._internal.preservation import Policy
from btrfs2s3._internal.preservation import Timeframe
from btrfs2s3._internal.preservation import TIMEFRAMES
from btrfs2s3._internal.zoneinfo import get_zoneinfo
import pytest


def _random_timestamp() -> float:
    return uniform(0.0, arrow.get("9999").timestamp())


@pytest.fixture(params=TIMEFRAMES)
def timeframe(request: pytest.FixtureRequest) -> Timeframe:
    return cast(Timeframe, request.param)


def test_empty_iter_time_spans(timeframe: Timeframe) -> None:
    policy = Policy(now=_random_timestamp())
    assert list(policy.iter_time_spans(_random_timestamp())) == []
    time_span = convert_span(
        arrow.get(_random_timestamp()).span(timeframe, bounds="[]")
    )
    assert not policy.should_preserve_for_time_span(time_span)


def test_empty_should_preserve_for_time_span(timeframe: Timeframe) -> None:
    policy = Policy(now=_random_timestamp())
    time_span = convert_span(
        arrow.get(_random_timestamp()).span(timeframe, bounds="[]")
    )
    assert not policy.should_preserve_for_time_span(time_span)


def test_all(timeframe: Timeframe) -> None:
    # This should be changed when we implement infinite preservation policies
    now = _random_timestamp()
    time_span = convert_span(arrow.get(now).span(timeframe, bounds="[]"))
    policy = Policy.all(now=now)
    assert time_span in list(policy.iter_time_spans(now))
    assert policy.should_preserve_for_time_span(time_span)


def test_some_time_frames() -> None:
    policy = Policy(params=Params(years=1))
    time_span = convert_span(arrow.get().span("year", bounds="[]"))

    start, end = time_span
    assert start < policy.now < end

    assert time_span in policy.iter_time_spans(time.time())
    assert policy.should_preserve_for_time_span(time_span)


def test_alternate_now() -> None:
    policy = Policy(now=0.0, params=Params(years=1))
    assert policy.now == 0.0
    time_span = convert_span(arrow.get(0).span("year", bounds="[]"))

    assert time_span in policy.iter_time_spans(0)
    assert policy.should_preserve_for_time_span(time_span)


def test_alternate_timezone() -> None:
    tzinfo = get_zoneinfo("America/Los_Angeles")
    policy = Policy(params=Params(years=1), tzinfo=tzinfo)
    time_span = convert_span(arrow.get(tzinfo=tzinfo).span("year", bounds="[]"))

    assert time_span in policy.iter_time_spans(time.time())
    assert policy.should_preserve_for_time_span(time_span)



"""Actions that modify snapshots or backups."""

from __future__ import annotations

import dataclasses
import errno
from itertools import chain
import logging
from subprocess import PIPE
from subprocess import Popen
from typing import TYPE_CHECKING

import btrfsutil

from btrfs2s3._internal.stream_uploader import upload_non_seekable_stream_via_tempfile
from btrfs2s3._internal.thunk import Thunk
from btrfs2s3._internal.thunk import ThunkArg
from btrfs2s3._internal.util import NULL_UUID
from btrfs2s3._internal.util import SubvolumeFlags

if TYPE_CHECKING:
    from pathlib import Path
    from typing import Iterator
    from typing import Sequence

    from mypy_boto3_s3.client import S3Client

_LOG = logging.getLogger(__name__)

DEFAULT_PART_SIZE = 5 * 2**30


def create_snapshot(*, source: Path, path: Path) -> None:
    """Create a read-only snapshot of a subvolume.

    Args:
        source: The source subvolume.
        path: The path at which to create a read-only snapshot.
    """
    _LOG.info("creating read-only snapshot of %s at %s", source, path)
    btrfsutil.create_snapshot(source, path, read_only=True)


def delete_snapshot(path: Path) -> None:
    """Delete a read-only snapshot of a subvolume.

    Args:
        path: The path to a read-only snapshot to be deleted.

    Raises:
        RuntimeError: If one of the arguments does not refer to a read-only
            snapshot of a subvolume.
    """
    # Do some extra checks to make sure we only ever delete read-only
    # snapshots, not source subvolumes.
    if not btrfsutil.is_subvolume(path):
        msg = "target isn't a subvolume"
        raise RuntimeError(msg)
    info = btrfsutil.subvolume_info(path)
    if info.parent_uuid == NULL_UUID:
        msg = "target isn't a snapshot"
        raise RuntimeError(msg)
    if not info.flags & SubvolumeFlags.ReadOnly:
        msg = "target isn't a read-only snapshot"
        raise RuntimeError(msg)
    _LOG.info("deleting read-only snapshot %s", path)
    try:
        btrfsutil.delete_subvolume(path)
    except OSError as ex:
        if ex.errno == errno.EROFS:
            # https://github.com/sbrudenell/btrfs2s3/issues/38
            # If we set the subvolume read-write but fail to delete it, later
            # runs will ignore it, creating a leak. To minimize this, only do
            # it in the non-root case.
            btrfsutil.set_subvolume_read_only(path, read_only=False)
            btrfsutil.delete_subvolume(path)
        else:
            raise  # pragma: no cover


def rename_snapshot(*, source: Path, target: Path) -> None:
    """Rename a read-only snapshot of of a subvolume.

    Args:
        source: The source snapshot.
        target: The new path of the snapshot.
    """
    _LOG.info("renaming %s -> %s", source, target)
    source.rename(target)


def create_backup(
    *,
    s3: S3Client,
    bucket: str,
    snapshot: Path,
    send_parent: Path | None,
    key: str,
    pipe_through: Sequence[Sequence[str]] = (),
    part_size: int = DEFAULT_PART_SIZE,
) -> None:
    """Stores a btrfs archive in S3.

    This will spawn "btrfs -q send" as a subprocess, as there is currently no way
    to create a btrfs-send stream via pure python.

    This will temporarily store the "btrfs send" stream on disk. A maximum of
    part_size bytes will be stored at a time.

    Args:
        s3: An S3 client.
        bucket: The bucket in which to store the archive.
        snapshot: The snapshot to back up.
        send_parent: The parent snapshot for delta backups. When not None, this
            will be supplied to "btrfs send -p".
        key: The S3 object key.
        pipe_through: A sequence of shell commands through which the archive
            should be piped before uploading.
        part_size: For multipart uploads, use this as the maximum part size.
            Defaults to the well-known maximum part size for AWS, which is
            currently 5 GiB.
    """
    _LOG.info(
        "creating backup of %s (%s)",
        snapshot,
        f"delta from {send_parent}" if send_parent else "full",
    )
    send_args: list[str | Path] = ["btrfs", "send", "-q"]
    if send_parent is not None:
        send_args += ["-p", send_parent]
    send_args += [snapshot]

    pipeline: list[Popen[bytes]] = []
    for args in chain((send_args,), pipe_through):
        prev_stdout = pipeline[-1].stdout if pipeline else None
        pipeline.append(Popen(args, stdin=prev_stdout, stdout=PIPE))  # noqa: S603
        # https://docs.python.org/3/library/subprocess.html#replacing-shell-pipeline
        if prev_stdout:
            prev_stdout.close()

    pipeline_stdout = pipeline[-1].stdout
    # https://github.com/python/typeshed/issues/3831
    assert pipeline_stdout is not None  # noqa: S101
    try:
        upload_non_seekable_stream_via_tempfile(
            client=s3,
            bucket=bucket,
            key=key,
            stream=pipeline_stdout,
            part_size=part_size,
        )
    finally:
        # Allow the pipeline to fail if the upload fails
        pipeline_stdout.close()

    # reverse order to match the semantics of pipefail
    for process in reversed(pipeline):
        if process.wait() != 0:
            msg = f"{process.args!r}: exited with code {process.returncode}"
            try:
                raise RuntimeError(msg)
            finally:
                # Assume the backup is corrupted
                delete_backups(s3, bucket, key)


def delete_backups(s3: S3Client, bucket: str, *keys: str) -> None:
    """Batch delete backups from S3.

    This will use the DeleteObjects API call, which can delete multiple keys in
    batches.

    Args:
        s3: An S3 client.
        bucket: The bucket from which to delete keys.
        *keys: The keys to delete.
    """

    def batches() -> Iterator[tuple[str, ...]]:
        for i in range(0, len(keys), 1000):
            yield keys[i : i + 1000]

    for batch in batches():
        for key in batch:
            _LOG.info("deleting backup %s", key)
        # Do we need to inspect the response for individual errors, or will we
        # raise an exception in this case? The docs are thousands of words long
        # but don't explain this
        s3.delete_objects(
            Bucket=bucket,
            Delete={"Quiet": True, "Objects": [{"Key": key} for key in batch]},
        )


# Future: all the fields of the intent objects should really be
# descriptor-typed fields that convert things to Thunk.


@dataclasses.dataclass(frozen=True, order=True)
class CreateSnapshot:
    """An intent to create a read-only snapshot of a subvolume."""

    source: Thunk[Path]
    path: Thunk[Path]


@dataclasses.dataclass(frozen=True, order=True)
class DeleteSnapshot:
    """An intent to delete a read-only snapshot."""

    path: Thunk[Path]


@dataclasses.dataclass(frozen=True, order=True)
class RenameSnapshot:
    """An intent to rename a read-only snapshot."""

    source: Thunk[Path]
    target: Thunk[Path]


@dataclasses.dataclass(frozen=True, order=True)
class CreateBackup:
    """An intent to create a backup of a read-only snapshot."""

    source: Thunk[Path]
    snapshot: Thunk[Path]
    send_parent: Thunk[Path | None]
    key: Thunk[str]


@dataclasses.dataclass(frozen=True, order=True)
class DeleteBackup:
    """An intent to delete a backup."""

    key: Thunk[str]


class Actions:
    """A list of actions to be executed.

    This class just stores a list of intents. Intents of actions can be added
    with the various functions, and then examined with the iter_*_intents()
    functions. It doesn't actually perform any actions until execute() is called.

    All of the arguments to the intent functions are converted to Thunks.

    Note that when creating Thunks, it's critical to understand the order that
    actions are performed. See execute().
    """

    def __init__(self) -> None:
        """Construct an empty Actions object."""
        self._create_snapshots: list[CreateSnapshot] = []
        self._delete_snapshots: list[DeleteSnapshot] = []
        self._rename_snapshots: list[RenameSnapshot] = []
        self._create_backups: list[CreateBackup] = []
        self._delete_backups: list[DeleteBackup] = []

    def empty(self) -> bool:
        """Returns True if there are no intended actions."""
        return not (
            self._create_snapshots
            or self._delete_snapshots
            or self._rename_snapshots
            or self._create_backups
            or self._delete_backups
        )

    def create_snapshot(self, *, source: ThunkArg[Path], path: ThunkArg[Path]) -> None:
        """Add an intent to create a read-only snapshot of a subvolume.

        Args:
            source: The path to the source subvolume, of which a snapshot
                should be created.
            path: The path where the new snapshot should be created.
        """
        self._create_snapshots.append(
            CreateSnapshot(source=Thunk(source), path=Thunk(path))
        )

    def rename_snapshot(
        self, *, source: ThunkArg[Path], target: ThunkArg[Path]
    ) -> None:
        """Add an intent to rename a snapshot.

        Args:
            source: The initial path of the snapshot.
            target: The new path of the snapshot.
        """
        self._rename_snapshots.append(
            RenameSnapshot(source=Thunk(source), target=Thunk(target))
        )

    def delete_snapshot(self, path: ThunkArg[Path]) -> None:
        """Add an intent to delete a snapshot.

        Args:
            path: The path of the snapshot to be deleted.
        """
        self._delete_snapshots.append(DeleteSnapshot(path=Thunk(path)))

    def create_backup(
        self,
        *,
        source: ThunkArg[Path],
        snapshot: ThunkArg[Path],
        send_parent: ThunkArg[Path | None],
        key: ThunkArg[str],
    ) -> None:
        """Add an intent to create a new backup of a read-only snapshot.

        The backup will be created using `btrfs send` (if send_parent is not
        None, then `btrfs send -p` will be used). The backup will be streamed
        into an S3 object with the given key.

        Args:
            source: The path to the source subvolume of the snapshot. This
                has no effect when creating the backup, but it's very useful
                when examining intents with iter_create_backup_intents().
            snapshot: The path to the read-only snapshot to be backed up.
            send_parent: The path to a read-only snapshot to be used as a
                parent (will be passed to `btrfs send -p`).
            key: The S3 object key at which to store the backup.
        """
        self._create_backups.append(
            CreateBackup(
                source=Thunk(source),
                snapshot=Thunk(snapshot),
                send_parent=Thunk(send_parent),
                key=Thunk(key),
            )
        )

    def delete_backup(self, key: ThunkArg[str]) -> None:
        """Add an intent to delete an S3 object.

        The target S3 object should ostensibly be a backup previously created,
        but we don't verify this.

        Args:
            key: The S3 object key to be deleted.
        """
        self._delete_backups.append(DeleteBackup(Thunk(key)))

    def iter_create_snapshot_intents(self) -> Iterator[CreateSnapshot]:
        """Iterates all the CreateSnapshot intents."""
        yield from sorted(self._create_snapshots)

    def iter_delete_snapshot_intents(self) -> Iterator[DeleteSnapshot]:
        """Iterates all the DeleteSnapshot intents."""
        yield from sorted(self._delete_snapshots)

    def iter_rename_snapshot_intents(self) -> Iterator[RenameSnapshot]:
        """Iterates all the RenameSnapshot intents."""
        yield from sorted(self._rename_snapshots)

    def iter_create_backup_intents(self) -> Iterator[CreateBackup]:
        """Iterates all the CreateBackup intents."""
        yield from sorted(self._create_backups)

    def iter_delete_backup_intents(self) -> Iterator[DeleteBackup]:
        """Iterates all the DeleteBackup intents."""
        yield from sorted(self._delete_backups)

    def execute(
        self, s3: S3Client, bucket: str, pipe_through: Sequence[Sequence[str]] = ()
    ) -> None:
        """Executes the intended actions.

        This performs all the side effects described in the various intent
        objects. It will create/rename/delete snapshots using btrfsutil, and
        create/delete backups in the supplied S3 bucket.

        The actions are performed in the following order:

        - Create snapshots
        - Rename snapshots
        - Create backups
        - Delete snapshots
        - Delete backups

        Within an action type, the actions are performed in the same order
        returned from the iter_*_intents() functions. That is, snapshots are
        created in the same order that they are returned from
        iter_create_snapshot_intents(), etc.

        Args:
            s3: The S3 client object to use to manipulate S3 objects.
            bucket: The name of the bucket where backups are stored.
            pipe_through: A sequence of shell commands through which backup
                archives should be piped before uploading.
        """
        for create_snapshot_intent in self.iter_create_snapshot_intents():
            create_snapshot(
                source=create_snapshot_intent.source(),
                path=create_snapshot_intent.path(),
            )

        for rename_snapshot_intent in self.iter_rename_snapshot_intents():
            rename_snapshot(
                source=rename_snapshot_intent.source(),
                target=rename_snapshot_intent.target(),
            )

        for create_backup_intent in self.iter_create_backup_intents():
            create_backup(
                s3=s3,
                bucket=bucket,
                snapshot=create_backup_intent.snapshot(),
                send_parent=create_backup_intent.send_parent(),
                key=create_backup_intent.key(),
                pipe_through=pipe_through,
            )

        for delete_snapshot_intent in self.iter_delete_snapshot_intents():
            delete_snapshot(delete_snapshot_intent.path())

        keys = tuple(d.key() for d in self.iter_delete_backup_intents())
        delete_backups(s3, bucket, *keys)
